"""
Object Store plugin for the Integrated Rule-Oriented Data Store (iRODS)

The module is named rods to avoid conflicting with the PyRods module, irods
"""

import os
import time
import logging

from posixpath import join as path_join
from posixpath import basename as path_basename
from posixpath import dirname as path_dirname

from galaxy.exceptions import ObjectNotFound
from ..objectstore import DiskObjectStore, ObjectStore, local_extra_dirs

try:
    import irods
except ImportError:
    irods = None


IRODS_IMPORT_MESSAGE = ('The Python irods package is required to use this '
                        'feature, please install it')

log = logging.getLogger( __name__ )


class IRODSObjectStore( DiskObjectStore, ObjectStore ):
    """
    Galaxy object store based on iRODS
    """
    def __init__( self, config, file_path=None, extra_dirs=None ):
        super( IRODSObjectStore, self ).__init__( config, file_path=file_path, extra_dirs=extra_dirs )
        assert irods is not None, IRODS_IMPORT_MESSAGE
        self.cache_path = config.object_store_cache_path
        self.default_resource = config.irods_default_resource or None

        # Connect to iRODS (AssertionErrors will be raised if anything goes wrong)
        self.rods_env, self.rods_conn = rods_connect()

        # if the root collection path in the config is unset or relative, try to use a sensible default
        if config.irods_root_collection_path is None or ( config.irods_root_collection_path is not None and not config.irods_root_collection_path.startswith( '/' ) ):
            rods_home = self.rods_env.rodsHome
            assert rods_home != '', "Unable to initialize iRODS Object Store: rodsHome cannot be determined and irods_root_collection_path in Galaxy config is unset or not absolute."
            if config.irods_root_collection_path is None:
                self.root_collection_path = path_join( rods_home, 'galaxy_data' )
            else:
                self.root_collection_path = path_join( rods_home, config.irods_root_collection_path )
        else:
            self.root_collection_path = config.irods_root_collection_path

        # will return a collection object regardless of whether it exists
        self.root_collection = irods.irodsCollection( self.rods_conn, self.root_collection_path )

        if self.root_collection.getId() == -1:
            log.warning( "iRODS root collection does not exist, will attempt to create: %s", self.root_collection_path )
            self.root_collection.upCollection()
            assert self.root_collection.createCollection( os.path.basename( self.root_collection_path ) ) == 0, "iRODS root collection creation failed: %s" % self.root_collection_path
            self.root_collection = irods.irodsCollection( self.rods_conn, self.root_collection_path )
            assert self.root_collection.getId() != -1, "iRODS root collection creation claimed success but still does not exist"

        if self.default_resource is None:
            self.default_resource = self.rods_env.rodsDefResource

        log.info( "iRODS data for this instance will be stored in collection: %s, resource: %s", self.root_collection_path, self.default_resource )

    def __get_rods_path( self, obj, base_dir=None, dir_only=False, extra_dir=None, extra_dir_at_root=False, alt_name=None, strip_dat=True, **kwargs ):
        path = ""
        if extra_dir is not None:
            path = extra_dir

        # extra_dir_at_root is ignored - since the iRODS plugin does not use
        # the directory hash, there is only one level of subdirectory.

        if not dir_only:
            # the .dat extension is stripped when stored in iRODS
            # TODO: is the strip_dat kwarg the best way to implement this?
            if strip_dat and alt_name and alt_name.endswith( '.dat' ):
                alt_name = os.path.splitext( alt_name )[0]
            default_name = 'dataset_%s' % obj.id
            if not strip_dat:
                default_name += '.dat'
            path = path_join( path, alt_name if alt_name else default_name )

        path = path_join( self.root_collection_path, path )
        return path

    def __get_cache_path( self, obj, **kwargs ):
        # FIXME: does not handle collections
        # FIXME: collisions could occur here
        return os.path.join( self.cache_path, path_basename( self.__get_rods_path( obj, strip_dat=False, **kwargs ) ) )

    def __clean_cache_entry( self, obj, **kwargs ):
        # FIXME: does not handle collections
        try:
            os.unlink( self.__get_cache_path( obj, **kwargs ) )
        except OSError:
            # it is expected that we'll call this method a lot regardless of
            # whether we think the cached file exists
            pass

    def __get_rods_handle( self, obj, mode='r', **kwargs ):
        if kwargs.get( 'dir_only', False ):
            return irods.irodsCollection( self.rods_conn, self.__get_rods_path( obj, **kwargs ) )
        else:
            return irods.irodsOpen( self.rods_conn, self.__get_rods_path( obj, **kwargs ), mode )

    def __mkcolls( self, rods_path ):
        """
        An os.makedirs() for iRODS collections.  `rods_path` is the desired collection to create.
        """
        assert rods_path.startswith( self.root_collection_path + '/' ), '__mkcolls(): Creating collections outside the root collection is not allowed (requested path was: %s)' % rods_path
        mkcolls = []
        c = irods.irodsCollection( self.rods_conn, rods_path )
        while c.getId() == -1:
            assert c.getCollName().startswith( self.root_collection_path + '/' ), '__mkcolls(): Attempted to move above the root collection: %s' % c.getCollName()
            mkcolls.append( c.getCollName() )
            c.upCollection()
        for collname in reversed( mkcolls ):
            log.debug( 'Creating collection %s' % collname )
            ci = irods.collInp_t()
            ci.collName = collname
            status = irods.rcCollCreate( self.rods_conn, ci )
            assert status == 0, '__mkcolls(): Failed to create collection: %s' % collname

    @local_extra_dirs
    def exists( self, obj, **kwargs ):
        doi = irods.dataObjInp_t()
        doi.objPath = self.__get_rods_path( obj, **kwargs )
        log.debug( 'exists(): checking: %s', doi.objPath )
        return irods.rcObjStat( self.rods_conn, doi ) is not None

    @local_extra_dirs
    def create(self, obj, **kwargs):
        if not self.exists( obj, **kwargs ):
            rods_path = self.__get_rods_path( obj, **kwargs )
            log.debug( 'create(): %s', rods_path )
            dir_only = kwargs.get( 'dir_only', False )
            # short circuit collection creation since most of the time it will
            # be the root collection which already exists
            collection_path = rods_path if dir_only else path_dirname( rods_path )
            if collection_path != self.root_collection_path:
                self.__mkcolls( collection_path )
            if not dir_only:
                # rcDataObjCreate is used instead of the irodsOpen wrapper so
                # that we can prevent overwriting
                doi = irods.dataObjInp_t()
                doi.objPath = rods_path
                doi.createMode = 0o640
                doi.dataSize = 0  # 0 actually means "unknown", although literally 0 would be preferable
                irods.addKeyVal( doi.condInput, irods.DEST_RESC_NAME_KW, self.default_resource )
                status = irods.rcDataObjCreate( self.rods_conn, doi )
                assert status >= 0, 'create(): rcDataObjCreate() failed: %s: %s: %s' % ( rods_path, status, irods.strerror( status ) )

    @local_extra_dirs
    def empty( self, obj, **kwargs ):
        assert 'dir_only' not in kwargs, 'empty(): `dir_only` parameter is invalid here'
        h = self.__get_rods_handle( obj, **kwargs )
        try:
            return h.getSize() == 0
        except AttributeError:
            # h is None
            raise ObjectNotFound()

    def size( self, obj, **kwargs ):
        assert 'dir_only' not in kwargs, 'size(): `dir_only` parameter is invalid here'
        h = self.__get_rods_handle( obj, **kwargs )
        try:
            return h.getSize()
        except AttributeError:
            # h is None
            return 0

    @local_extra_dirs
    def delete( self, obj, entire_dir=False, **kwargs ):
        assert 'dir_only' not in kwargs, 'delete(): `dir_only` parameter is invalid here'
        rods_path = self.__get_rods_path( obj, **kwargs )
        # __get_rods_path prepends self.root_collection_path but we are going
        # to ensure that it's valid anyway for safety's sake
        assert rods_path.startswith( self.root_collection_path + '/' ), 'ERROR: attempt to delete object outside root collection (path was: %s)' % rods_path
        if entire_dir:
            # TODO
            raise NotImplementedError()
        h = self.__get_rods_handle( obj, **kwargs )
        try:
            # note: PyRods' irodsFile.delete() does not set force
            status = h.delete()
            assert status == 0, '%d: %s' % ( status, irods.strerror( status ) )
            return True
        except AttributeError:
            log.warning( 'delete(): operation failed: object does not exist: %s', rods_path )
        except AssertionError as e:
            # delete() does not raise on deletion failure
            log.error( 'delete(): operation failed: %s', e )
        finally:
            # remove the cached entry (finally is executed even when the try
            # contains a return)
            self.__clean_cache_entry( self, obj, **kwargs )
        return False

    @local_extra_dirs
    def get_data( self, obj, start=0, count=-1, **kwargs ):
        log.debug( 'get_data(): %s' )
        h = self.__get_rods_handle( obj, **kwargs )
        try:
            h.seek( start )
        except AttributeError:
            raise ObjectNotFound()
        if count == -1:
            return h.read()
        else:
            return h.read( count )
        # TODO: make sure implicit close is okay, DiskObjectStore actually
        # reads data into a var, closes, and returns the var

    @local_extra_dirs
    def get_filename( self, obj, **kwargs ):
        log.debug( "get_filename(): called on %s %s. For better performance, avoid this method and use get_data() instead.", obj.__class__.__name__, obj.id )
        cached_path = self.__get_cache_path( obj, **kwargs )

        if not self.exists( obj, **kwargs ):
            raise ObjectNotFound()

        # TODO: implement or define whether dir_only is valid
        if 'dir_only' in kwargs:
            raise NotImplementedError()

        # cache hit
        if os.path.exists( cached_path ):
            return os.path.abspath( cached_path )

        # cache miss
        # TODO: thread this
        incoming_path = os.path.join( os.path.dirname( cached_path ), "__incoming_%s" % os.path.basename( cached_path ) )
        doi = irods.dataObjInp_t()
        doi.objPath = self.__get_rods_path( obj, **kwargs )
        doi.dataSize = 0  # TODO: does this affect performance? should we get size?
        doi.numThreads = 0
        # TODO: might want to VERIFY_CHKSUM_KW
        log.debug( 'get_filename(): caching %s to %s', doi.objPath, incoming_path )

        # do the iget
        status = irods.rcDataObjGet( self.rods_conn, doi, incoming_path )

        # if incoming already exists, we'll wait for another process or thread
        # to finish caching
        if status != irods.OVERWRITE_WITHOUT_FORCE_FLAG:
            assert status == 0, 'get_filename(): iget %s failed (%s): %s' % ( doi.objPath, status, irods.strerror( status ) )
            # POSIX rename is atomic
            # TODO: rename without clobbering
            os.rename( incoming_path, cached_path )
            log.debug( 'get_filename(): cached %s to %s', doi.objPath, cached_path )

        # another process or thread is caching, wait for it
        while not os.path.exists( cached_path ):
            # TODO: force restart after mod time > some configurable, or
            # otherwise deal with this potential deadlock and interrupted
            # transfers
            time.sleep( 5 )
            log.debug( "get_filename(): waiting on incoming '%s' for %s %s", incoming_path, obj.__class__.__name__, obj.id )

        return os.path.abspath( cached_path )

    @local_extra_dirs
    def update_from_file(self, obj, file_name=None, create=False, **kwargs):
        assert 'dir_only' not in kwargs, 'update_from_file(): `dir_only` parameter is invalid here'

        # do not create if not requested
        if create and not self.exists( obj, **kwargs ):
            raise ObjectNotFound()

        if file_name is None:
            file_name = self.__get_cache_path( obj, **kwargs )

        # put will create if necessary
        doi = irods.dataObjInp_t()
        doi.objPath = self.__get_rods_path( obj, **kwargs )
        doi.createMode = 0o640
        doi.dataSize = os.stat( file_name ).st_size
        doi.numThreads = 0
        irods.addKeyVal( doi.condInput, irods.DEST_RESC_NAME_KW, self.default_resource )
        irods.addKeyVal( doi.condInput, irods.FORCE_FLAG_KW, '' )
        # TODO: might want to VERIFY_CHKSUM_KW
        log.debug( 'update_from_file(): updating %s to %s', file_name, doi.objPath )

        # do the iput
        status = irods.rcDataObjPut( self.rods_conn, doi, file_name )
        assert status == 0, 'update_from_file(): iput %s failed (%s): %s' % ( doi.objPath, status, irods.strerror( status ) )

    def get_object_url(self, obj, **kwargs):
        return None

    def get_store_usage_percent(self):
        return 0.0


# monkeypatch an strerror method into the irods module
def _rods_strerror( errno ):
    """
    The missing `strerror` for iRODS error codes
    """
    if not hasattr( irods, '__rods_strerror_map' ):
        irods.__rods_strerror_map = {}
        for name in dir( irods ):
            v = getattr( irods, name )
            if type( v ) == int and v < 0:
                irods.__rods_strerror_map[ v ] = name
    return irods.__rods_strerror_map.get( errno, 'GALAXY_NO_ERRNO_MAPPING_FOUND' )

if irods is not None:
    irods.strerror = _rods_strerror


def rods_connect():
    """
    A basic iRODS connection mechanism that connects using the current iRODS
    environment
    """
    status, env = irods.getRodsEnv()
    assert status == 0, 'connect(): getRodsEnv() failed (%s): %s' % ( status, irods.strerror( status ) )
    conn, err = irods.rcConnect( env.rodsHost,
                                 env.rodsPort,
                                 env.rodsUserName,
                                 env.rodsZone )
    assert err.status == 0, 'connect(): rcConnect() failed (%s): %s' % ( err.status, err.msg )
    status, pw = irods.obfGetPw()
    assert status == 0, 'connect(): getting password with obfGetPw() failed (%s): %s' % ( status, irods.strerror( status ) )
    status = irods.clientLoginWithObfPassword( conn, pw )
    assert status == 0, 'connect(): logging in with clientLoginWithObfPassword() failed (%s): %s' % ( status, irods.strerror( status ) )
    return env, conn



from ..lint_util import is_datasource


def lint_inputs(tool_xml, lint_ctx):
    datasource = is_datasource(tool_xml)
    inputs = tool_xml.findall("./inputs//param")
    num_inputs = 0
    for param in inputs:
        num_inputs += 1
        param_attrib = param.attrib
        has_errors = False
        if "type" not in param_attrib:
            lint_ctx.error("Found param input with no type specified.")
            has_errors = True
        if "name" not in param_attrib and "argument" not in param_attrib:
            lint_ctx.error("Found param input with no name specified.")
            has_errors = True

        if has_errors:
            continue

        param_type = param_attrib["type"]
        param_name = param_attrib.get("name", param_attrib.get("argument"))
        if param_type == "data":
            if "format" not in param_attrib:
                lint_ctx.warn("Param input [%s] with no format specified - 'data' format will be assumed.", param_name)

        if param_type == "select":
            dynamic_options = param.get("dynamic_options", None)
            if dynamic_options is None:
                dynamic_options = param.find("options")

            select_options = _find_with_attribute(param, 'option', 'value')
            if any(['value' not in option.attrib for option in select_options]):
                lint_ctx.error("Option without value")

            select_option_ids = [option.attrib.get('value', None) for option in select_options]

            if dynamic_options is None and len(select_options) == 0:
                message = "No options defined for select [%s]" % param_name
                lint_ctx.warn(message)

        # TODO: Validate type, much more...

    conditional_selects = tool_xml.findall("./inputs//conditional")
    for conditional in conditional_selects:
        booleans = _find_with_attribute(conditional, "param", "type", "boolean")
        selects = _find_with_attribute(conditional, "param", "type", "select")
        # Should conditionals ever not have a select?
        if not len(selects) and not len(booleans):
            lint_ctx.warn("Conditional without <param type=\"select\" /> or <param type=\"boolean\" />")
            continue

        for select in selects:
            select_options = _find_with_attribute(select, 'option', 'value')
            select_option_ids = [option.attrib.get('value', None) for option in select_options]

        for boolean in booleans:
            select_option_ids = [
                boolean.attrib.get('truevalue', 'true'),
                boolean.attrib.get('falsevalue', 'false')
            ]

        whens = conditional.findall('./when')
        if any(['value' not in when.attrib for when in whens]):
            lint_ctx.error("When without value")

        when_ids = [w.attrib.get('value', None) for w in whens]
        when_ids = [i.lower() if i in ["True", "False"] else i for i in when_ids]

        for select_id in select_option_ids:
            if select_id not in when_ids:
                lint_ctx.warn("No <when /> block found for select option '%s'" % select_id)

        for when_id in when_ids:
            if when_id not in select_option_ids:
                lint_ctx.warn("No <option /> block found for when block '%s'" % when_id)

    if datasource:
        for datasource_tag in ('display', 'uihints'):
            if not any([param.tag == datasource_tag for param in inputs]):
                lint_ctx.info("%s tag usually present in data sources" % datasource_tag)

    if num_inputs:
        lint_ctx.info("Found %d input parameters.", num_inputs)
    else:
        if datasource:
            lint_ctx.info("No input parameters, OK for data sources")
        else:
            lint_ctx.warn("Found no input parameters.")


def lint_repeats(tool_xml, lint_ctx):
    repeats = tool_xml.findall("./inputs//repeat")
    for repeat in repeats:
        if "name" not in repeat.attrib:
            lint_ctx.error("Repeat does not specify name attribute.")
        if "title" not in repeat.attrib:
            lint_ctx.error("Repeat does not specify title attribute.")


def _find_with_attribute(element, tag, attribute, test_value=None):
    rval = []
    for el in (element.findall('./%s' % tag) or []):
        if attribute not in el.attrib:
            continue
        value = el.attrib[attribute]
        if test_value is not None:
            if value == test_value:
                rval.append(el)
        else:
            rval.append(el)
    return rval



import os.path
try:
    from watchdog.events import FileSystemEventHandler
    from watchdog.observers import Observer
    from watchdog.observers.polling import PollingObserver
    can_watch = True
except ImportError:
    FileSystemEventHandler = object
    PollingObserver = object
    can_watch = False

import logging
log = logging.getLogger( __name__ )


def get_watcher(toolbox, config):
    watch_tools_val = str( getattr(config, "watch_tools", False) ).lower()
    if watch_tools_val in ( 'true', 'yes', 'on' ):
        return ToolWatcher(toolbox)
    elif watch_tools_val == "auto":
        try:
            return ToolWatcher(toolbox)
        except Exception:
            log.info("Failed to load ToolWatcher (watchdog is likely unavailable) - proceeding without tool monitoring.")
            return NullWatcher()
    elif watch_tools_val == "polling":
        log.info("Using less ineffecient polling toolbox watcher.")
        return ToolWatcher(toolbox, observer_class=PollingObserver)
    else:
        return NullWatcher()


class ToolWatcher(object):

    def __init__(self, toolbox, observer_class=None):
        if not can_watch:
            raise Exception("Watchdog library unavailble, cannot watch tools.")
        if observer_class is None:
            observer_class = Observer
        self.toolbox = toolbox
        self.tool_file_ids = {}
        self.tool_dir_callbacks = {}
        self.monitored_dirs = {}
        self.observer = observer_class()
        self.event_handler = ToolFileEventHandler(self)
        self.start()

    def start(self):
        self.observer.start()

    def monitor(self, dir):
        self.observer.schedule(self.event_handler, dir, recursive=False)

    def watch_file(self, tool_file, tool_id):
        tool_file = os.path.abspath( tool_file )
        self.tool_file_ids[tool_file] = tool_id
        tool_dir = os.path.dirname( tool_file )
        if tool_dir not in self.monitored_dirs:
            self.monitored_dirs[ tool_dir ] = tool_dir
            self.monitor( tool_dir )

    def watch_directory(self, tool_dir, callback):
        tool_dir = os.path.abspath( tool_dir )
        self.tool_dir_callbacks[tool_dir] = callback
        if tool_dir not in self.monitored_dirs:
            self.monitored_dirs[ tool_dir ] = tool_dir
            self.monitor( tool_dir )


class ToolFileEventHandler(FileSystemEventHandler):

    def __init__(self, tool_watcher):
        self.tool_watcher = tool_watcher

    def on_any_event(self, event):
        self._handle(event)

    def _handle(self, event):
        # modified events will only have src path, move events will
        # have dest_path and src_path but we only care about dest. So
        # look at dest if it exists else use src.
        path = getattr( event, 'dest_path', None ) or event.src_path
        path = os.path.abspath( path )
        tool_id = self.tool_watcher.tool_file_ids.get( path, None )
        if tool_id:
            try:
                self.tool_watcher.toolbox.reload_tool_by_id(tool_id)
            except Exception:
                pass
        elif path.endswith(".xml"):
            directory = os.path.dirname( path )
            dir_callback = self.tool_watcher.tool_dir_callbacks.get( directory, None )
            if dir_callback:
                tool_file = event.src_path
                tool_id = dir_callback( tool_file )
                if tool_id:
                    self.tool_watcher.tool_file_ids[ tool_file ] = tool_id


class NullWatcher(object):

    def start(self):
        pass

    def shutdown(self):
        pass

    def watch_file(self, tool_file, tool_id):
        pass

    def watch_directory(self, tool_dir, callback):
        pass



from abc import ABCMeta, abstractmethod


class DependencyResolver( object ):
    __metaclass__ = ABCMeta

    @abstractmethod
    def resolve( self, name, version, type, **kwds ):
        """
        Given inputs describing dependency in the abstract, yield tuple of
        (script, bin, version). Here script is the env.sh file to source
        before running a job, if that is not found the bin directory will be
        appended to the path (if it is not None). Finally, version is the
        resolved tool dependency version (which may differ from requested
        version for instance if the request version is 'default'.)
        """

    def _get_config_option(self, key, dependency_resolver, default=None, prefix=None, **kwds):
        """ Look in resolver-specific settings for option and then fallback to
        global settings.
        """
        global_key = "%s_%s" % (prefix, key)
        if key in kwds:
            return kwds.get(key)
        elif global_key in dependency_resolver.extra_config:
            return dependency_resolver.extra_config.get(global_key)
        else:
            return default


class Dependency( object ):
    __metaclass__ = ABCMeta

    @abstractmethod
    def shell_commands( self, requirement ):
        """
        Return shell commands to enable this dependency.
        """


class NullDependency( Dependency ):

    def shell_commands( self, requirement ):
        return None

INDETERMINATE_DEPENDENCY = NullDependency()



from galaxy.util.xml_macros import load, raw_tool_xml_tree, imported_macro_paths, template_macro_params

load_tool = load

__all__ = ["load_tool", "raw_tool_xml_tree", "imported_macro_paths", "template_macro_params"]





def directory_hash_id( id ):
    """

    >>> directory_hash_id( 100 )
    ['000']
    >>> directory_hash_id( "90000" )
    ['090']
    >>> directory_hash_id("777777777")
    ['000', '777', '777']
    """
    s = str( id )
    l = len( s )
    # Shortcut -- ids 0-999 go under ../000/
    if l < 4:
        return [ "000" ]
    # Pad with zeros until a multiple of three
    padded = ( ( 3 - len( s ) % 3 ) * "0" ) + s
    # Drop the last three digits -- 1000 files per directory
    padded = padded[:-3]
    # Break into chunks of three
    return [ padded[ i * 3 : (i + 1 ) * 3 ] for i in range( len( padded ) // 3 ) ]



from .command import get_command


def lint_stdio(tool_xml, lint_ctx):
    stdios = tool_xml.findall("./stdio")
    if not stdios:
        command = get_command(tool_xml)
        if command is None or not command.get("detect_errors"):
            lint_ctx.info("No stdio definition found, tool will determine an error from stderr.")
        return

    if len(stdios) > 1:
        lint_ctx.error("More than one stdio tag found, behavior undefined.")
        return

    stdio = stdios[0]
    for child in list(stdio):
        if child.tag == "regex":
            _lint_regex(child, lint_ctx)
        elif child.tag == "exit_code":
            _lint_exit_code(child, lint_ctx)
        else:
            message = "Unknown stdio child tag discovered [%s]. "
            message += "Valid options are exit_code and regex."
            lint_ctx.warn(message % child.tag)


def _lint_exit_code(child, lint_ctx):
    for key, value in child.attrib.items():
        if key == "range":
            # TODO: validate
            pass
        elif key == "level":
            _lint_level(value, lint_ctx)
        elif key == "description":
            pass
        else:
            lint_ctx.warn("Unknown attribute [%s] encountered on exit_code tag." % key)


def _lint_regex(child, lint_ctx):
    for key, value in child.attrib.iteritems():
        if key == "source":
            if value not in ["stderr", "stdout", "both"]:
                lint_ctx.error("Unknown error code level encountered [%s]" % value)
        elif key == "level":
            _lint_level(value, lint_ctx)
        elif key == "match":
            # TODO: validate
            pass
        elif key == "description":
            pass
        else:
            lint_ctx.warn("Unknown attribute [%s] encountered on regex tag." % key)


def _lint_level(level_value, lint_ctx):
    if level_value not in ["warning", "fatal", "log"]:
        lint_ctx.error("Unknown error code level encountered [%s]" % level_value)



from __future__ import print_function
import inspect
from galaxy.util import submodules

LEVEL_ALL = "all"
LEVEL_WARN = "warn"
LEVEL_ERROR = "error"


def lint_xml(tool_xml, level=LEVEL_ALL, fail_level=LEVEL_WARN, extra_modules=[], skip_types=[]):
    lint_context = LintContext(level=level, skip_types=skip_types)
    lint_xml_with(lint_context, tool_xml, extra_modules)

    return not lint_context.failed(fail_level)


def lint_xml_with(lint_context, tool_xml, extra_modules=[]):
    import galaxy.tools.linters
    linter_modules = submodules.submodules(galaxy.tools.linters)
    linter_modules.extend(extra_modules)
    for module in linter_modules:
        for (name, value) in inspect.getmembers(module):
            if callable(value) and name.startswith("lint_"):
                lint_context.lint(name, value, tool_xml)


class LintContext(object):

    def __init__(self, level, skip_types=[]):
        self.skip_types = skip_types
        self.level = level
        self.found_errors = False
        self.found_warns = False

    def lint(self, name, lint_func, lint_target):
        name = name.replace("tsts", "tests")[len("lint_"):]
        if name in self.skip_types:
            return
        self.printed_linter_info = False
        self.valid_messages = []
        self.info_messages = []
        self.warn_messages = []
        self.error_messages = []
        lint_func(lint_target, self)
        # TODO: colorful emoji if in click CLI.
        if self.error_messages:
            status = "FAIL"
        elif self.warn_messages:
            status = "WARNING"
        else:
            status = "CHECK"

        def print_linter_info():
            if self.printed_linter_info:
                return
            self.printed_linter_info = True
            print("Applying linter %s... %s" % (name, status))

        for message in self.error_messages:
            self.found_errors = True
            print_linter_info()
            print(".. ERROR: %s" % message)

        if self.level != LEVEL_ERROR:
            for message in self.warn_messages:
                self.found_warns = True
                print_linter_info()
                print(".. WARNING: %s" % message)

        if self.level == LEVEL_ALL:
            for message in self.info_messages:
                print_linter_info()
                print(".. INFO: %s" % message)
            for message in self.valid_messages:
                print_linter_info()
                print(".. CHECK: %s" % message)

    def __handle_message(self, message_list, message, *args):
        if args:
            message = message % args
        message_list.append(message)

    def valid(self, message, *args):
        self.__handle_message(self.valid_messages, message, *args)

    def info(self, message, *args):
        self.__handle_message(self.info_messages, message, *args)

    def error(self, message, *args):
        self.__handle_message(self.error_messages, message, *args)

    def warn(self, message, *args):
        self.__handle_message(self.warn_messages, message, *args)

    def failed(self, fail_level):
        found_warns = self.found_warns
        found_errors = self.found_errors
        if fail_level == LEVEL_WARN:
            lint_fail = (found_warns or found_errors)
        else:
            lint_fail = found_errors
        return lint_fail



""" This module contains helper functions and data structures for interacting
with collectl and collectl generated data. More information on collectl can be
found at: http://collectl.sourceforge.net/.

"""



import collections
import os

from galaxy.util import plugin_config
from galaxy import util

from ..metrics import formatting

import logging
log = logging.getLogger( __name__ )


DEFAULT_FORMATTER = formatting.JobMetricFormatter()


class JobMetrics( object ):

    def __init__( self, conf_file=None, **kwargs ):
        """
        """
        self.plugin_classes = self.__plugins_dict()
        self.default_job_instrumenter = JobInstrumenter.from_file( self.plugin_classes, conf_file, **kwargs )
        self.job_instrumenters = collections.defaultdict( lambda: self.default_job_instrumenter )

    def format( self, plugin, key, value ):
        if plugin in self.plugin_classes:
            plugin_class = self.plugin_classes[ plugin ]
            formatter = plugin_class.formatter
        else:
            formatter = DEFAULT_FORMATTER
        return formatter.format( key, value )

    def set_destination_conf_file( self, destination_id, conf_file ):
        instrumenter = JobInstrumenter.from_file( self.plugin_classes, conf_file )
        self.set_destination_instrumenter( destination_id, instrumenter )

    def set_destination_conf_element( self, destination_id, element ):
        instrumenter = JobInstrumenter( self.plugin_classes, ('xml', element) )
        self.set_destination_instrumenter( destination_id, instrumenter )

    def set_destination_instrumenter( self, destination_id, job_instrumenter=None ):
        if job_instrumenter is None:
            job_instrumenter = NULL_JOB_INSTRUMENTER
        self.job_instrumenters[ destination_id ] = job_instrumenter

    def collect_properties( self, destination_id, job_id, job_directory ):
        return self.job_instrumenters[ destination_id ].collect_properties( job_id, job_directory )

    def __plugins_dict( self ):
        import galaxy.jobs.metrics.instrumenters
        return plugin_config.plugins_dict( galaxy.jobs.metrics.instrumenters, 'plugin_type' )


class NullJobInstrumenter( object ):

    def pre_execute_commands( self, job_directory ):
        return None

    def post_execute_commands( self, job_directory ):
        return None

    def collect_properties( self, job_id, job_directory ):
        return {}

NULL_JOB_INSTRUMENTER = NullJobInstrumenter()


class JobInstrumenter( object ):

    def __init__( self, plugin_classes, plugins_source, **kwargs ):
        self.extra_kwargs = kwargs
        self.plugin_classes = plugin_classes
        self.plugins = self.__plugins_from_source( plugins_source )

    def pre_execute_commands( self, job_directory ):
        commands = []
        for plugin in self.plugins:
            try:
                plugin_commands = plugin.pre_execute_instrument( job_directory )
                if plugin_commands:
                    commands.extend( util.listify( plugin_commands ) )
            except Exception:
                log.exception( "Failed to generate pre-execute commands for plugin %s" % plugin )
        return "\n".join( [ c for c in commands if c ] )

    def post_execute_commands( self, job_directory ):
        commands = []
        for plugin in self.plugins:
            try:
                plugin_commands = plugin.post_execute_instrument( job_directory )
                if plugin_commands:
                    commands.extend( util.listify( plugin_commands ) )
            except Exception:
                log.exception( "Failed to generate post-execute commands for plugin %s" % plugin )
        return "\n".join( [ c for c in commands if c ] )

    def collect_properties( self, job_id, job_directory ):
        per_plugin_properites = {}
        for plugin in self.plugins:
            try:
                properties = plugin.job_properties( job_id, job_directory )
                if properties:
                    per_plugin_properites[ plugin.plugin_type ] = properties
            except Exception:
                log.exception( "Failed to collect job properties for plugin %s" % plugin )
        return per_plugin_properites

    def __plugins_from_source( self, plugins_source ):
        return plugin_config.load_plugins(self.plugin_classes, plugins_source, self.extra_kwargs)

    @staticmethod
    def from_file( plugin_classes, conf_file, **kwargs ):
        if not conf_file or not os.path.exists( conf_file ):
            return NULL_JOB_INSTRUMENTER
        plugins_source = plugin_config.plugin_source_from_path( conf_file )
        return JobInstrumenter( plugin_classes, plugins_source, **kwargs )



from .interface import ToolLineage
from .interface import ToolLineageVersion

try:
    from galaxy.model.tool_shed_install import ToolVersion
except ImportError:
    ToolVersion = None


class ToolShedLineage(ToolLineage):
    """ Representation of tool lineage derived from tool shed repository
    installations. """

    def __init__(self, app, tool_version, tool_shed_repository=None):
        if ToolVersion is None:
            raise Exception("Tool shed models not present, can't create tool shed lineages.")
        self.app = app
        self.tool_version_id = tool_version.id
        # Only used for logging
        self._tool_shed_repository = tool_shed_repository

    @staticmethod
    def from_tool( app, tool, tool_shed_repository ):
        # Make sure the tool has a tool_version.
        if not get_install_tool_version( app, tool.id ):
            tool_version = ToolVersion( tool_id=tool.id, tool_shed_repository=tool_shed_repository )
            app.install_model.context.add( tool_version )
            app.install_model.context.flush()
        return ToolShedLineage( app, tool.tool_version )

    @staticmethod
    def from_tool_id( app, tool_id ):
        tool_version = get_install_tool_version( app, tool_id )
        if tool_version:
            return ToolShedLineage( app, tool_version )
        else:
            return None

    def get_version_ids( self, reverse=False ):
        tool_version = self.app.install_model.context.query( ToolVersion ).get( self.tool_version_id )
        return tool_version.get_version_ids( self.app, reverse=reverse )

    def get_versions( self, reverse=False ):
        return map( ToolLineageVersion.from_guid, self.get_version_ids( reverse=reverse ) )

    def to_dict(self):
        tool_shed_repository = self._tool_shed_repository
        rval = dict(
            tool_version_id=self.tool_version_id,
            tool_versions=map(lambda v: v.to_dict(), self.get_versions()),
            tool_shed_repository=tool_shed_repository if tool_shed_repository is not None else None,
            lineage_type='tool_shed',
        )
        return rval


def get_install_tool_version( app, tool_id ):
    return app.install_model.context.query(
        app.install_model.ToolVersion
    ).filter(
        app.install_model.ToolVersion.table.c.tool_id == tool_id
    ).first()

__all__ = [ "ToolShedLineage" ]



from .tool_shed import ToolShedLineage
from .stock import StockLineage


class LineageMap(object):
    """ Map each unique tool id to a lineage object.
    """

    def __init__(self, app):
        self.lineage_map = {}
        self.app = app

    def register(self, tool, **kwds):
        tool_id = tool.id
        if tool_id not in self.lineage_map:
            tool_shed_repository = kwds.get("tool_shed_repository", None)
            if tool_shed_repository:
                lineage = ToolShedLineage.from_tool(self.app, tool, tool_shed_repository)
            else:
                lineage = StockLineage.from_tool( tool )
            self.lineage_map[tool_id] = lineage
        return self.lineage_map[tool_id]

    def get(self, tool_id):
        if tool_id not in self.lineage_map:
            lineage = ToolShedLineage.from_tool_id( self.app, tool_id )
            if lineage:
                self.lineage_map[tool_id] = lineage

        return self.lineage_map.get(tool_id, None)

__all__ = ["LineageMap"]



import os
import subprocess
import sys as _sys


def redirecting_io(sys=_sys):
    assert sys is not None
    # We are redirecting standard out and standard error.
    return not hasattr(sys.stdout, "fileno")


def redirect_aware_commmunicate(p, sys=_sys):
    assert sys is not None
    out, err = p.communicate()
    if redirecting_io(sys=sys):
        if out:
            sys.stdout.write(out)
            out = None
        if err:
            sys.stderr.write(err)
            err = None
    return out, err


def shell(cmds, env=None, **kwds):
    sys = kwds.get("sys", _sys)
    assert sys is not None
    p = shell_process(cmds, env, **kwds)
    if redirecting_io(sys=sys):
        redirect_aware_commmunicate(p, sys=sys)
        exit = p.returncode
        return exit
    else:
        return p.wait()


def shell_process(cmds, env=None, **kwds):
    sys = kwds.get("sys", _sys)
    popen_kwds = dict(
        shell=True,
    )
    if kwds.get("stdout", None) is None and redirecting_io(sys=sys):
        popen_kwds["stdout"] = subprocess.PIPE
    if kwds.get("stderr", None) is None and redirecting_io(sys=sys):
        popen_kwds["stderr"] = subprocess.PIPE

    popen_kwds.update(**kwds)
    if env:
        new_env = os.environ.copy()
        new_env.update(env)
        popen_kwds["env"] = new_env
    p = subprocess.Popen(cmds, **popen_kwds)
    return p


def execute(cmds):
    return __wait(cmds, shell=False)


def which(file):
    # http://stackoverflow.com/questions/5226958/which-equivalent-function-in-python
    for path in os.environ["PATH"].split(":"):
        if os.path.exists(path + "/" + file):
                return path + "/" + file

    return None


def __wait(cmds, **popen_kwds):
    p = subprocess.Popen(cmds, **popen_kwds)
    stdout, stderr = p.communicate()
    if p.returncode != 0:
        raise CommandLineException(" ".join(cmds), stdout, stderr)
    return stdout


def download_command(url, quote_url=False):
    if quote_url:
        url = "'%s'" % url
    if which("wget"):
        download_cmd = ["wget", "-q", "--recursive", "-O" "-", url]
    else:
        download_cmd = ["curl", "-L", url]
    return download_cmd


class CommandLineException(Exception):

    def __init__(self, command, stdout, stderr):
        self.command = command
        self.stdout = stdout
        self.stderr = stderr
        self.message = ("Failed to execute command-line %s, stderr was:\n"
                        "-------->>begin stderr<<--------\n"
                        "%s\n"
                        "-------->>end stderr<<--------\n"
                        "-------->>begin stdout<<--------\n"
                        "%s\n"
                        "-------->>end stdout<<--------\n"
                        ) % (command, stderr, stdout)

    def __str__(self):
        return self.message



"""
This is still an experimental module and there will almost certainly be backward
incompatible changes coming.
"""


import os

from ..resolvers import DependencyResolver, INDETERMINATE_DEPENDENCY
from ..conda_util import (
    CondaContext,
    CondaTarget,
    install_conda,
    is_conda_target_installed,
    install_conda_target,
    build_isolated_environment,
)

DEFAULT_ENSURE_CHANNELS = "r,bioconda"

import logging
log = logging.getLogger(__name__)


class CondaDependencyResolver(DependencyResolver):
    resolver_type = "conda"

    def __init__(self, dependency_manager, **kwds):
        self.versionless = _string_as_bool(kwds.get('versionless', 'false'))

        def get_option(name):
            return self._get_config_option(name, dependency_manager, prefix="conda", **kwds)

        # Conda context options (these define the environment)
        conda_prefix = get_option("prefix")
        conda_exec = get_option("exec")
        debug = _string_as_bool(get_option("debug"))
        ensure_channels = get_option("ensure_channels")
        if ensure_channels is None:
            ensure_channels = DEFAULT_ENSURE_CHANNELS

        conda_context = CondaContext(
            conda_prefix=conda_prefix,
            conda_exec=conda_exec,
            debug=debug,
            ensure_channels=ensure_channels,
        )

        # Conda operations options (these define how resolution will occur)
        auto_init = _string_as_bool(get_option("auto_init"))
        auto_install = _string_as_bool(get_option("auto_install"))
        copy_dependencies = _string_as_bool(get_option("copy_dependencies"))

        if auto_init and not os.path.exists(conda_context.conda_prefix):
            install_conda(conda_context)

        self.conda_context = conda_context
        self.auto_install = auto_install
        self.copy_dependencies = copy_dependencies

    def resolve(self, name, version, type, **kwds):
        # Check for conda just not being there, this way we can enable
        # conda by default and just do nothing in not configured.
        if not os.path.isdir(self.conda_context.conda_prefix):
            return INDETERMINATE_DEPENDENCY

        if type != "package":
            return INDETERMINATE_DEPENDENCY

        job_directory = kwds.get("job_directory", None)
        if job_directory is None:
            log.warn("Conda dependency resolver not sent job directory.")
            return INDETERMINATE_DEPENDENCY

        if self.versionless:
            version = None

        conda_target = CondaTarget(name, version=version)
        is_installed = is_conda_target_installed(
            conda_target, conda_context=self.conda_context
        )
        if not is_installed and self.auto_install:
            install_conda_target(conda_target)

            # Recheck if installed
            is_installed = is_conda_target_installed(
                conda_target, conda_context=self.conda_context
            )

        if not is_installed:
            return INDETERMINATE_DEPENDENCY

        # Have installed conda_target and job_directory to send it too.
        conda_environment = os.path.join(job_directory, "conda-env")
        env_path, exit_code = build_isolated_environment(
            conda_target,
            path=conda_environment,
            copy=self.copy_dependencies,
            conda_context=self.conda_context,
        )
        if not exit_code:
            return CondaDepenency(
                self.conda_context.activate,
                conda_environment
            )
        else:
            raise Exception("Conda dependency seemingly installed but failed to build job environment.")


class CondaDepenency():

    def __init__(self, activate, environment_path):
        self.activate = activate
        self.environment_path = environment_path

    def shell_commands(self, requirement):
        return """[ "$CONDA_DEFAULT_ENV" = "%s" ] || source %s '%s'""" % (
            self.environment_path,
            self.activate,
            self.environment_path
        )


def _string_as_bool( value ):
    return str( value ).lower() == "true"


__all__ = ['CondaDependencyResolver']



import logging
import os
import re
import string
import tarfile
import tempfile

from markupsafe import escape
from six.moves.urllib.parse import urlparse

from galaxy.util.dictifiable import Dictifiable

from galaxy.util.odict import odict
from galaxy.util import listify
from galaxy.util import parse_xml
from galaxy.util import string_as_bool
from galaxy.util.bunch import Bunch

from .panel import ToolPanelElements
from .panel import ToolSectionLabel
from .panel import ToolSection
from .panel import panel_item_types
from .integrated_panel import ManagesIntegratedToolPanelMixin

from .lineages import LineageMap
from .tags import tool_tag_manager

from .filters import FilterFactory
from .watcher import get_watcher


log = logging.getLogger( __name__ )


class AbstractToolBox( Dictifiable, ManagesIntegratedToolPanelMixin, object ):
    """
    Abstract container for managing a ToolPanel - containing tools and
    workflows optionally in labelled sections.
    """

    def __init__( self, config_filenames, tool_root_dir, app ):
        """
        Create a toolbox from the config files named by `config_filenames`, using
        `tool_root_dir` as the base directory for finding individual tool config files.
        """
        # The _dynamic_tool_confs list contains dictionaries storing
        # information about the tools defined in each shed-related
        # shed_tool_conf.xml file.
        self._dynamic_tool_confs = []
        self._tools_by_id = {}
        self._integrated_section_by_tool = {}
        # Tool lineages can contain chains of related tools with different ids
        # so each will be present once in the above dictionary. The following
        # dictionary can instead hold multiple tools with different versions.
        self._tool_versions_by_id = {}
        self._workflows_by_id = {}
        # In-memory dictionary that defines the layout of the tool panel.
        self._tool_panel = ToolPanelElements()
        self._index = 0
        self.data_manager_tools = odict()
        self._lineage_map = LineageMap( app )
        # Sets self._integrated_tool_panel and self._integrated_tool_panel_config_has_contents
        self._init_integrated_tool_panel( app.config )
        # The following refers to the tool_path config setting for backward compatibility.  The shed-related
        # (e.g., shed_tool_conf.xml) files include the tool_path attribute within the <toolbox> tag.
        self._tool_root_dir = tool_root_dir
        self.app = app
        self._tool_watcher = get_watcher( self, app.config )
        self._filter_factory = FilterFactory( self )
        self._tool_tag_manager = tool_tag_manager( app )
        self._init_tools_from_configs( config_filenames )
        if self.app.name == 'galaxy' and self._integrated_tool_panel_config_has_contents:
            # Load self._tool_panel based on the order in self._integrated_tool_panel.
            self._load_tool_panel()
        self._save_integrated_tool_panel()

    def create_tool( self, config_file, repository_id=None, guid=None, **kwds ):
        raise NotImplementedError()

    def _init_tools_from_configs( self, config_filenames ):
        """ Read through all tool config files and initialize tools in each
        with init_tools_from_config below.
        """
        self._tool_tag_manager.reset_tags()
        config_filenames = listify( config_filenames )
        for config_filename in config_filenames:
            if os.path.isdir( config_filename ):
                directory_contents = sorted( os.listdir( config_filename ) )
                directory_config_files = [ config_file for config_file in directory_contents if config_file.endswith( ".xml" ) ]
                config_filenames.remove( config_filename )
                config_filenames.extend( directory_config_files )
        for config_filename in config_filenames:
            try:
                self._init_tools_from_config( config_filename )
            except:
                log.exception( "Error loading tools defined in config %s", config_filename )

    def _init_tools_from_config( self, config_filename ):
        """
        Read the configuration file and load each tool.  The following tags are currently supported:

        .. raw:: xml

            <toolbox>
                <tool file="data_source/upload.xml"/>                 # tools outside sections
                <label text="Basic Tools" id="basic_tools" />         # labels outside sections
                <workflow id="529fd61ab1c6cc36" />                    # workflows outside sections
                <section name="Get Data" id="getext">                 # sections
                    <tool file="data_source/biomart.xml" />           # tools inside sections
                    <label text="In Section" id="in_section" />       # labels inside sections
                    <workflow id="adb5f5c93f827949" />                # workflows inside sections
                    <tool file="data_source/foo.xml" labels="beta" /> # label for a single tool
                </section>
            </toolbox>

        """
        log.info( "Parsing the tool configuration %s" % config_filename )
        tree = parse_xml( config_filename )
        root = tree.getroot()
        tool_path = root.get( 'tool_path' )
        if tool_path:
            # We're parsing a shed_tool_conf file since we have a tool_path attribute.
            parsing_shed_tool_conf = True
            # Keep an in-memory list of xml elements to enable persistence of the changing tool config.
            config_elems = []
        else:
            parsing_shed_tool_conf = False
        tool_path = self.__resolve_tool_path(tool_path, config_filename)
        # Only load the panel_dict under certain conditions.
        load_panel_dict = not self._integrated_tool_panel_config_has_contents
        for _, elem in enumerate( root ):
            index = self._index
            self._index += 1
            if parsing_shed_tool_conf:
                config_elems.append( elem )
            self.load_item(
                elem,
                tool_path=tool_path,
                load_panel_dict=load_panel_dict,
                guid=elem.get( 'guid' ),
                index=index,
                internal=True
            )

        if parsing_shed_tool_conf:
            shed_tool_conf_dict = dict( config_filename=config_filename,
                                        tool_path=tool_path,
                                        config_elems=config_elems )
            self._dynamic_tool_confs.append( shed_tool_conf_dict )

    def load_item( self, elem, tool_path, panel_dict=None, integrated_panel_dict=None, load_panel_dict=True, guid=None, index=None, internal=False ):
        item_type = elem.tag
        if item_type not in ['tool', 'section'] and not internal:
            # External calls from tool shed code cannot load labels or tool
            # directories.
            return

        if panel_dict is None:
            panel_dict = self._tool_panel
        if integrated_panel_dict is None:
            integrated_panel_dict = self._integrated_tool_panel
        if item_type == 'tool':
            self._load_tool_tag_set( elem, panel_dict=panel_dict, integrated_panel_dict=integrated_panel_dict, tool_path=tool_path, load_panel_dict=load_panel_dict, guid=guid, index=index )
        elif item_type == 'workflow':
            self._load_workflow_tag_set( elem, panel_dict=panel_dict, integrated_panel_dict=integrated_panel_dict, load_panel_dict=load_panel_dict, index=index )
        elif item_type == 'section':
            self._load_section_tag_set( elem, tool_path=tool_path, load_panel_dict=load_panel_dict, index=index )
        elif item_type == 'label':
            self._load_label_tag_set( elem, panel_dict=panel_dict, integrated_panel_dict=integrated_panel_dict, load_panel_dict=load_panel_dict, index=index )
        elif item_type == 'tool_dir':
            self._load_tooldir_tag_set( elem, panel_dict, tool_path, integrated_panel_dict, load_panel_dict=load_panel_dict )

    def get_shed_config_dict_by_filename( self, filename, default=None ):
        for shed_config_dict in self._dynamic_tool_confs:
            if shed_config_dict[ 'config_filename' ] == filename:
                return shed_config_dict
        return default

    def update_shed_config( self, shed_conf, integrated_panel_changes=True ):
        """ Update the in-memory descriptions of tools and write out the changes
        to integrated tool panel unless we are just deactivating a tool (since
        that doesn't affect that file).
        """
        app = self.app
        for index, my_shed_tool_conf in enumerate( self._dynamic_tool_confs ):
            if shed_conf['config_filename'] == my_shed_tool_conf['config_filename']:
                self._dynamic_tool_confs[ index ] = shed_conf
        if integrated_panel_changes:
            self._save_integrated_tool_panel()
        app.reindex_tool_search()

    def get_section( self, section_id, new_label=None, create_if_needed=False ):
        tool_panel_section_key = str( section_id )
        if tool_panel_section_key in self._tool_panel:
            # Appending a tool to an existing section in toolbox._tool_panel
            tool_section = self._tool_panel[ tool_panel_section_key ]
            log.debug( "Appending to tool panel section: %s" % str( tool_section.name ) )
        elif create_if_needed:
            # Appending a new section to toolbox._tool_panel
            if new_label is None:
                # This might add an ugly section label to the tool panel, but, oh well...
                new_label = section_id
            section_dict = {
                'name': new_label,
                'id': section_id,
                'version': '',
            }
            tool_section = ToolSection( section_dict )
            self._tool_panel.append_section( tool_panel_section_key, tool_section )
            log.debug( "Loading new tool panel section: %s" % str( tool_section.name ) )
        else:
            tool_section = None
        return tool_panel_section_key, tool_section

    def get_integrated_section_for_tool( self, tool ):
        tool_id = tool.id

        if tool_id in self._integrated_section_by_tool:
            return self._integrated_section_by_tool[tool_id]

        return None, None

    def __resolve_tool_path(self, tool_path, config_filename):
        if not tool_path:
            # Default to backward compatible config setting.
            tool_path = self._tool_root_dir
        else:
            # Allow use of __tool_conf_dir__ in toolbox config files.
            tool_conf_dir = os.path.dirname(config_filename)
            tool_path_vars = {"tool_conf_dir": tool_conf_dir}
            tool_path = string.Template(tool_path).safe_substitute(tool_path_vars)
        return tool_path

    def __add_tool_to_tool_panel( self, tool, panel_component, section=False ):
        # See if a version of this tool is already loaded into the tool panel.
        # The value of panel_component will be a ToolSection (if the value of
        # section=True) or self._tool_panel (if section=False).
        tool_id = str( tool.id )
        tool = self._tools_by_id[ tool_id ]
        if section:
            panel_dict = panel_component.elems
        else:
            panel_dict = panel_component

        related_tool = self._lineage_in_panel( panel_dict, tool=tool )
        if related_tool:
            if self._newer_tool( tool, related_tool ):
                panel_dict.replace_tool(
                    previous_tool_id=related_tool.id,
                    new_tool_id=tool_id,
                    tool=tool,
                )
                log.debug( "Loaded tool id: %s, version: %s into tool panel." % ( tool.id, tool.version ) )
        else:
            inserted = False
            index = self._integrated_tool_panel.index_of_tool_id( tool_id )
            if index:
                panel_dict.insert_tool( index, tool )
                inserted = True
            if not inserted:
                # Check the tool's installed versions.
                versions = []
                if hasattr( tool, 'lineage' ):
                    versions = tool.lineage.get_versions()
                for tool_lineage_version in versions:
                    lineage_id = tool_lineage_version.id
                    index = self._integrated_tool_panel.index_of_tool_id( lineage_id )
                    if index:
                        panel_dict.insert_tool( index, tool )
                        inserted = True
                if not inserted:
                    if (
                        tool.guid is None or
                        tool.tool_shed is None or
                        tool.repository_name is None or
                        tool.repository_owner is None or
                        tool.installed_changeset_revision is None
                    ):
                        # We have a tool that was not installed from the Tool
                        # Shed, but is also not yet defined in
                        # integrated_tool_panel.xml, so append it to the tool
                        # panel.
                        panel_dict.append_tool( tool )
                        log.debug( "Loaded tool id: %s, version: %s into tool panel.." % ( tool.id, tool.version ) )
                    else:
                        # We are in the process of installing the tool.

                        tool_lineage = self._lineage_map.get( tool_id )
                        already_loaded = self._lineage_in_panel( panel_dict, tool_lineage=tool_lineage ) is not None
                        if not already_loaded:
                            # If the tool is not defined in integrated_tool_panel.xml, append it to the tool panel.
                            panel_dict.append_tool( tool )
                            log.debug( "Loaded tool id: %s, version: %s into tool panel...." % ( tool.id, tool.version ) )

    def _load_tool_panel( self ):
        for key, item_type, val in self._integrated_tool_panel.panel_items_iter():
            if item_type == panel_item_types.TOOL:
                tool_id = key.replace( 'tool_', '', 1 )
                if tool_id in self._tools_by_id:
                    self.__add_tool_to_tool_panel( val, self._tool_panel, section=False )
                    self._integrated_section_by_tool[tool_id] = '', ''
            elif item_type == panel_item_types.WORKFLOW:
                workflow_id = key.replace( 'workflow_', '', 1 )
                if workflow_id in self._workflows_by_id:
                    workflow = self._workflows_by_id[ workflow_id ]
                    self._tool_panel[ key ] = workflow
                    log.debug( "Loaded workflow: %s %s" % ( workflow_id, workflow.name ) )
            elif item_type == panel_item_types.LABEL:
                self._tool_panel[ key ] = val
            elif item_type == panel_item_types.SECTION:
                section_dict = {
                    'id': val.id or '',
                    'name': val.name or '',
                    'version': val.version or '',
                }
                section = ToolSection( section_dict )
                log.debug( "Loading section: %s" % section_dict.get( 'name' ) )
                for section_key, section_item_type, section_val in val.panel_items_iter():
                    if section_item_type == panel_item_types.TOOL:
                        tool_id = section_key.replace( 'tool_', '', 1 )
                        if tool_id in self._tools_by_id:
                            self.__add_tool_to_tool_panel( section_val, section, section=True )
                            self._integrated_section_by_tool[tool_id] = key, val.name
                    elif section_item_type == panel_item_types.WORKFLOW:
                        workflow_id = section_key.replace( 'workflow_', '', 1 )
                        if workflow_id in self._workflows_by_id:
                            workflow = self._workflows_by_id[ workflow_id ]
                            section.elems[ section_key ] = workflow
                            log.debug( "Loaded workflow: %s %s" % ( workflow_id, workflow.name ) )
                    elif section_item_type == panel_item_types.LABEL:
                        if section_val:
                            section.elems[ section_key ] = section_val
                            log.debug( "Loaded label: %s" % ( section_val.text ) )
                self._tool_panel[ key ] = section

    def _load_integrated_tool_panel_keys( self ):
        """
        Load the integrated tool panel keys, setting values for tools and
        workflows to None.  The values will be reset when the various tool
        panel config files are parsed, at which time the tools and workflows
        are loaded.
        """
        tree = parse_xml( self._integrated_tool_panel_config )
        root = tree.getroot()
        for elem in root:
            key = elem.get( 'id' )
            if elem.tag == 'tool':
                self._integrated_tool_panel.stub_tool( key )
            elif elem.tag == 'workflow':
                self._integrated_tool_panel.stub_workflow( key )
            elif elem.tag == 'section':
                section = ToolSection( elem )
                for section_elem in elem:
                    section_id = section_elem.get( 'id' )
                    if section_elem.tag == 'tool':
                        section.elems.stub_tool( section_id )
                    elif section_elem.tag == 'workflow':
                        section.elems.stub_workflow( section_id )
                    elif section_elem.tag == 'label':
                        section.elems.stub_label( section_id )
                self._integrated_tool_panel.append_section( key, section )
            elif elem.tag == 'label':
                self._integrated_tool_panel.stub_label( key )

    def get_tool( self, tool_id, tool_version=None, get_all_versions=False, exact=False ):
        """Attempt to locate a tool in the tool box."""
        if tool_version:
            tool_version = str( tool_version )

        if get_all_versions and exact:
            raise AssertionError("Cannot specify get_tool with both get_all_versions and exact as True")

        if "/repos/" in tool_id:  # test if tool came from a toolshed
            tool_id_without_tool_shed = tool_id.split("/repos/")[1]
            available_tool_sheds = self.app.tool_shed_registry.tool_sheds.values()
            available_tool_sheds = [ urlparse(tool_shed) for tool_shed in available_tool_sheds ]
            available_tool_sheds = [ url.geturl().replace(url.scheme + "://", '', 1) for url in available_tool_sheds]
            tool_ids = [ tool_shed + "repos/" + tool_id_without_tool_shed for tool_shed in available_tool_sheds]
            if tool_id in tool_ids:  # move original tool_id to the top of tool_ids
                tool_ids.remove(tool_id)
            tool_ids.insert(0, tool_id)
        else:
            tool_ids = [tool_id]
        for tool_id in tool_ids:
            if tool_id in self._tools_by_id and not get_all_versions:
                if tool_version and tool_version in self._tool_versions_by_id[ tool_id ]:
                    return self._tool_versions_by_id[ tool_id ][ tool_version ]
                # tool_id exactly matches an available tool by id (which is 'old' tool_id or guid)
                return self._tools_by_id[ tool_id ]
            # exact tool id match not found, or all versions requested, search for other options, e.g. migrated tools or different versions
            rval = []
            tool_lineage = self._lineage_map.get( tool_id )
            if tool_lineage:
                lineage_tool_versions = tool_lineage.get_versions( )
                for lineage_tool_version in lineage_tool_versions:
                    lineage_tool = self._tool_from_lineage_version( lineage_tool_version )
                    if lineage_tool:
                        rval.append( lineage_tool )
            if not rval:
                # still no tool, do a deeper search and try to match by old ids
                for tool in self._tools_by_id.itervalues():
                    if tool.old_id == tool_id:
                        rval.append( tool )
            if rval:
                if get_all_versions:
                    return rval
                else:
                    if tool_version:
                        # return first tool with matching version
                        for tool in rval:
                            if tool.version == tool_version:
                                return tool
                    # No tool matches by version, simply return the first available tool found
                    return rval[0]
            # We now likely have a Toolshed guid passed in, but no supporting database entries
            # If the tool exists by exact id and is loaded then provide exact match within a list
            if tool_id in self._tools_by_id:
                return[ self._tools_by_id[ tool_id ] ]
        return None

    def has_tool( self, tool_id, tool_version=None, exact=False ):
        return self.get_tool( tool_id, tool_version=tool_version, exact=exact ) is not None

    def get_tool_id( self, tool_id ):
        """ Take a tool id (potentially from a different Galaxy instance or that
        is no longer loaded  - and find the closest match to the currently loaded
        tools (using get_tool for inexact matches which currently returns the oldest
        tool shed installed tool with the same short id).
        """
        if tool_id not in self._tools_by_id:
            tool = self.get_tool( tool_id )
            if tool:
                tool_id = tool.id
            else:
                tool_id = None
        # else exact match - leave unmodified.
        return tool_id

    def get_loaded_tools_by_lineage( self, tool_id ):
        """Get all loaded tools associated by lineage to the tool whose id is tool_id."""
        tool_lineage = self._lineage_map.get( tool_id )
        if tool_lineage:
            lineage_tool_versions = tool_lineage.get_versions( )
            available_tool_versions = []
            for lineage_tool_version in lineage_tool_versions:
                tool = self._tool_from_lineage_version( lineage_tool_version )
                if tool:
                    available_tool_versions.append( tool )
            return available_tool_versions
        else:
            if tool_id in self._tools_by_id:
                tool = self._tools_by_id[ tool_id ]
                return [ tool ]
        return []

    def tools( self ):
        return self._tools_by_id.iteritems()

    def dynamic_confs( self, include_migrated_tool_conf=False ):
        confs = []
        for dynamic_tool_conf_dict in self._dynamic_tool_confs:
            dynamic_tool_conf_filename = dynamic_tool_conf_dict[ 'config_filename' ]
            if include_migrated_tool_conf or (dynamic_tool_conf_filename != self.app.config.migrated_tools_config):
                confs.append( dynamic_tool_conf_dict )
        return confs

    def dynamic_conf_filenames( self, include_migrated_tool_conf=False ):
        """ Return list of dynamic tool configuration filenames (shed_tools).
        These must be used with various dynamic tool configuration update
        operations (e.g. with update_shed_config).
        """
        for dynamic_tool_conf_dict in self.dynamic_confs( include_migrated_tool_conf=include_migrated_tool_conf ):
            yield dynamic_tool_conf_dict[ 'config_filename' ]

    def remove_from_panel( self, tool_id, section_key='', remove_from_config=True ):

        def remove_from_dict( has_elems, integrated_has_elems ):
            tool_key = 'tool_%s' % str( tool_id )
            available_tool_versions = self.get_loaded_tools_by_lineage( tool_id )
            if tool_key in has_elems:
                if available_tool_versions:
                    available_tool_versions.reverse()
                    replacement_tool_key = None
                    replacement_tool_version = None
                    # Since we are going to remove the tool from the section, replace it with
                    # the newest loaded version of the tool.
                    for available_tool_version in available_tool_versions:
                        available_tool_section_id, available_tool_section_name = available_tool_version.get_panel_section()
                        # I suspect "available_tool_version.id in has_elems.keys()" doesn't
                        # belong in the following line or at least I don't understand what
                        # purpose it might serve. -John
                        if available_tool_version.id in has_elems.keys() or (available_tool_section_id == section_key):
                            replacement_tool_key = 'tool_%s' % str( available_tool_version.id )
                            replacement_tool_version = available_tool_version
                            break
                    if replacement_tool_key and replacement_tool_version:
                        # Get the index of the tool_key in the tool_section.
                        for tool_panel_index, key in enumerate( has_elems.keys() ):
                            if key == tool_key:
                                break
                        # Remove the tool from the tool panel.
                        del has_elems[ tool_key ]
                        # Add the replacement tool at the same location in the tool panel.
                        has_elems.insert( tool_panel_index,
                                          replacement_tool_key,
                                          replacement_tool_version )
                        self._integrated_section_by_tool[ tool_id ] = available_tool_section_id, available_tool_section_name
                    else:
                        del has_elems[ tool_key ]

                        if tool_id in self._integrated_section_by_tool:
                            del self._integrated_section_by_tool[ tool_id ]
                else:
                    del has_elems[ tool_key ]

                    if tool_id in self._integrated_section_by_tool:
                        del self._integrated_section_by_tool[ tool_id ]
            if remove_from_config:
                itegrated_items = integrated_has_elems.panel_items()
                if tool_key in itegrated_items:
                    del itegrated_items[ tool_key ]

        if section_key:
            _, tool_section = self.get_section( section_key )
            if tool_section:
                remove_from_dict( tool_section.elems, self._integrated_tool_panel.get( section_key, {} ) )
        else:
            remove_from_dict( self._tool_panel, self._integrated_tool_panel )

    def _load_tool_tag_set( self, elem, panel_dict, integrated_panel_dict, tool_path, load_panel_dict, guid=None, index=None ):
        try:
            path = elem.get( "file" )
            repository_id = None

            tool_shed_repository = None
            can_load_into_panel_dict = True
            if guid is not None:
                # The tool is contained in an installed tool shed repository, so load
                # the tool only if the repository has not been marked deleted.
                tool_shed = elem.find( "tool_shed" ).text
                repository_name = elem.find( "repository_name" ).text
                repository_owner = elem.find( "repository_owner" ).text
                installed_changeset_revision_elem = elem.find( "installed_changeset_revision" )
                if installed_changeset_revision_elem is None:
                    # Backward compatibility issue - the tag used to be named 'changeset_revision'.
                    installed_changeset_revision_elem = elem.find( "changeset_revision" )
                installed_changeset_revision = installed_changeset_revision_elem.text
                tool_shed_repository = self._get_tool_shed_repository( tool_shed,
                                                                       repository_name,
                                                                       repository_owner,
                                                                       installed_changeset_revision )

                if tool_shed_repository:
                    # Only load tools if the repository is not deactivated or uninstalled.
                    can_load_into_panel_dict = not tool_shed_repository.deleted
                    repository_id = self.app.security.encode_id( tool_shed_repository.id )
                # Else there is not yet a tool_shed_repository record, we're in the process of installing
                # a new repository, so any included tools can be loaded into the tool panel.
            tool = self.load_tool( os.path.join( tool_path, path ), guid=guid, repository_id=repository_id )
            if string_as_bool(elem.get( 'hidden', False )):
                tool.hidden = True
            key = 'tool_%s' % str( tool.id )
            if can_load_into_panel_dict:
                if guid is not None:
                    tool.tool_shed = tool_shed
                    tool.repository_name = repository_name
                    tool.repository_owner = repository_owner
                    tool.installed_changeset_revision = installed_changeset_revision
                    tool.guid = guid
                    tool.version = elem.find( "version" ).text
                # Make sure the tool has a tool_version.
                tool_lineage = self._lineage_map.register( tool, tool_shed_repository=tool_shed_repository )
                # Load the tool's lineage ids.
                tool.lineage = tool_lineage
                self._tool_tag_manager.handle_tags( tool.id, elem )
                self.__add_tool( tool, load_panel_dict, panel_dict )
            # Always load the tool into the integrated_panel_dict, or it will not be included in the integrated_tool_panel.xml file.
            integrated_panel_dict.update_or_append( index, key, tool )
            # If labels were specified in the toolbox config, attach them to
            # the tool.
            if "labels" in elem.attrib:
                tool.labels = [ label.strip() for label in elem.attrib["labels"].split( "," ) ]
        except IOError:
            log.error( "Error reading tool configuration file from path: %s." % path )
        except Exception:
            log.exception( "Error reading tool from path: %s" % path )

    def _get_tool_shed_repository( self, tool_shed, name, owner, installed_changeset_revision ):
        # Abstract class does't have a dependency on the database, for full tool shed
        # support the actual Galaxy ToolBox implement this method and return a ToolShd repository.
        return None

    def __add_tool( self, tool, load_panel_dict, panel_dict ):
        # Allow for the same tool to be loaded into multiple places in the
        # tool panel.  We have to handle the case where the tool is contained
        # in a repository installed from the tool shed, and the Galaxy
        # administrator has retrieved updates to the installed repository.  In
        # this case, the tool may have been updated, but the version was not
        # changed, so the tool should always be reloaded here.  We used to
        # only load the tool if it was not found in self._tools_by_id, but
        # performing that check did not enable this scenario.
        self.register_tool( tool )
        if load_panel_dict:
            self.__add_tool_to_tool_panel( tool, panel_dict, section=isinstance( panel_dict, ToolSection ) )

    def _load_workflow_tag_set( self, elem, panel_dict, integrated_panel_dict, load_panel_dict, index=None ):
        try:
            # TODO: should id be encoded?
            workflow_id = elem.get( 'id' )
            workflow = self._load_workflow( workflow_id )
            self._workflows_by_id[ workflow_id ] = workflow
            key = 'workflow_' + workflow_id
            if load_panel_dict:
                panel_dict[ key ] = workflow
            # Always load workflows into the integrated_panel_dict.
            integrated_panel_dict.update_or_append( index, key, workflow )
        except:
            log.exception( "Error loading workflow: %s" % workflow_id )

    def _load_label_tag_set( self, elem, panel_dict, integrated_panel_dict, load_panel_dict, index=None ):
        label = ToolSectionLabel( elem )
        key = 'label_' + label.id
        if load_panel_dict:
            panel_dict[ key ] = label
        integrated_panel_dict.update_or_append( index, key, label )

    def _load_section_tag_set( self, elem, tool_path, load_panel_dict, index=None ):
        key = elem.get( "id" )
        if key in self._tool_panel:
            section = self._tool_panel[ key ]
            elems = section.elems
        else:
            section = ToolSection( elem )
            elems = section.elems
        if key in self._integrated_tool_panel:
            integrated_section = self._integrated_tool_panel[ key ]
            integrated_elems = integrated_section.elems
        else:
            integrated_section = ToolSection( elem )
            integrated_elems = integrated_section.elems
        for sub_index, sub_elem in enumerate( elem ):
            self.load_item(
                sub_elem,
                tool_path=tool_path,
                panel_dict=elems,
                integrated_panel_dict=integrated_elems,
                load_panel_dict=load_panel_dict,
                guid=sub_elem.get( 'guid' ),
                index=sub_index,
                internal=True,
            )

        # Ensure each tool's section is stored
        for section_key, section_item_type, section_item in integrated_elems.panel_items_iter():
            if section_item_type == panel_item_types.TOOL:
                if section_item:
                    tool_id = section_key.replace( 'tool_', '', 1 )
                    self._integrated_section_by_tool[tool_id] = integrated_section.id, integrated_section.name

        if load_panel_dict:
            self._tool_panel[ key ] = section
        # Always load sections into the integrated_tool_panel.
        self._integrated_tool_panel.update_or_append( index, key, integrated_section )

    def _load_tooldir_tag_set(self, sub_elem, elems, tool_path, integrated_elems, load_panel_dict):
        directory = os.path.join( tool_path, sub_elem.attrib.get("dir") )
        recursive = string_as_bool( sub_elem.attrib.get("recursive", True) )
        self.__watch_directory( directory, elems, integrated_elems, load_panel_dict, recursive, force_watch=True )

    def __watch_directory( self, directory, elems, integrated_elems, load_panel_dict, recursive, force_watch=False ):

        def quick_load( tool_file, async=True ):
            try:
                tool = self.load_tool( tool_file )
                self.__add_tool( tool, load_panel_dict, elems )
                # Always load the tool into the integrated_panel_dict, or it will not be included in the integrated_tool_panel.xml file.
                key = 'tool_%s' % str( tool.id )
                integrated_elems[ key ] = tool

                if async:
                    self._load_tool_panel()
                    self._save_integrated_tool_panel()
                return tool.id
            except Exception:
                log.exception("Failed to load potential tool %s." % tool_file)
                return None

        tool_loaded = False
        for name in os.listdir( directory ):
            child_path = os.path.join(directory, name)
            if os.path.isdir(child_path) and recursive:
                self.__watch_directory(child_path, elems, integrated_elems, load_panel_dict, recursive)
            elif name.endswith( ".xml" ):
                quick_load( child_path, async=False )
                tool_loaded = True
        if tool_loaded or force_watch:
            self._tool_watcher.watch_directory( directory, quick_load )

    def load_tool( self, config_file, guid=None, repository_id=None, **kwds ):
        """Load a single tool from the file named by `config_file` and return an instance of `Tool`."""
        # Parse XML configuration file and get the root element
        tool = self.create_tool( config_file=config_file, repository_id=repository_id, guid=guid, **kwds )
        tool_id = tool.id
        if not tool_id.startswith("__"):
            # do not monitor special tools written to tmp directory - no reason
            # to monitor such a large directory.
            self._tool_watcher.watch_file( config_file, tool.id )
        return tool

    def load_hidden_lib_tool( self, path ):
        tool_xml = os.path.join( os.getcwd(), "lib", path )
        return self.load_hidden_tool( tool_xml )

    def load_hidden_tool( self, config_file, **kwds ):
        """ Load a hidden tool (in this context meaning one that does not
        appear in the tool panel) and register it in _tools_by_id.
        """
        tool = self.load_tool( config_file, **kwds )
        self.register_tool( tool )
        return tool

    def register_tool( self, tool ):
        tool_id = tool.id
        version = tool.version or None
        if tool_id not in self._tool_versions_by_id:
            self._tool_versions_by_id[ tool_id ] = { version: tool }
        else:
            self._tool_versions_by_id[ tool_id ][ version ] = tool
        if tool_id in self._tools_by_id:
            related_tool = self._tools_by_id[ tool_id ]
            # This one becomes the default un-versioned tool
            # if newer.
            if self._newer_tool( tool, related_tool ):
                self._tools_by_id[ tool_id ] = tool
        else:
            self._tools_by_id[ tool_id ] = tool

    def package_tool( self, trans, tool_id ):
        """
        Create a tarball with the tool's xml, help images, and test data.
        :param trans: the web transaction
        :param tool_id: the tool ID from app.toolbox
        :returns: tuple of tarball filename, success True/False, message/None
        """
        # Make sure the tool is actually loaded.
        if tool_id not in self._tools_by_id:
            return None, False, "No tool with id %s" % escape( tool_id )
        else:
            tool = self._tools_by_id[ tool_id ]
            tarball_files = []
            temp_files = []
            tool_xml = open( os.path.abspath( tool.config_file ), 'r' ).read()
            # Retrieve tool help images and rewrite the tool's xml into a temporary file with the path
            # modified to be relative to the repository root.
            image_found = False
            if tool.help is not None:
                tool_help = tool.help._source
                # Check each line of the rendered tool help for an image tag that points to a location under static/
                for help_line in tool_help.split( '\n' ):
                    image_regex = re.compile( 'img alt="[^"]+" src="\${static_path}/([^"]+)"' )
                    matches = re.search( image_regex, help_line )
                    if matches is not None:
                        tool_help_image = matches.group(1)
                        tarball_path = tool_help_image
                        filesystem_path = os.path.abspath( os.path.join( trans.app.config.root, 'static', tool_help_image ) )
                        if os.path.exists( filesystem_path ):
                            tarball_files.append( ( filesystem_path, tarball_path ) )
                            image_found = True
                            tool_xml = tool_xml.replace( '${static_path}/%s' % tarball_path, tarball_path )
            # If one or more tool help images were found, add the modified tool XML to the tarball instead of the original.
            if image_found:
                fd, new_tool_config = tempfile.mkstemp( suffix='.xml' )
                os.close( fd )
                open( new_tool_config, 'w' ).write( tool_xml )
                tool_tup = ( os.path.abspath( new_tool_config ), os.path.split( tool.config_file )[-1]  )
                temp_files.append( os.path.abspath( new_tool_config ) )
            else:
                tool_tup = ( os.path.abspath( tool.config_file ), os.path.split( tool.config_file )[-1]  )
            tarball_files.append( tool_tup )
            # TODO: This feels hacky.
            tool_command = tool.command.strip().split()[0]
            tool_path = os.path.dirname( os.path.abspath( tool.config_file ) )
            # Add the tool XML to the tuple that will be used to populate the tarball.
            if os.path.exists( os.path.join( tool_path, tool_command ) ):
                tarball_files.append( ( os.path.join( tool_path, tool_command ), tool_command ) )
            # Find and add macros and code files.
            for external_file in tool.get_externally_referenced_paths( os.path.abspath( tool.config_file ) ):
                external_file_abspath = os.path.abspath( os.path.join( tool_path, external_file ) )
                tarball_files.append( ( external_file_abspath, external_file ) )
            if os.path.exists( os.path.join( tool_path, "Dockerfile" ) ):
                tarball_files.append( ( os.path.join( tool_path, "Dockerfile" ), "Dockerfile" ) )
            # Find tests, and check them for test data.
            tests = tool.tests
            if tests is not None:
                for test in tests:
                    # Add input file tuples to the list.
                    for input in test.inputs:
                        for input_value in test.inputs[ input ]:
                            input_path = os.path.abspath( os.path.join( 'test-data', input_value ) )
                            if os.path.exists( input_path ):
                                td_tup = ( input_path, os.path.join( 'test-data', input_value ) )
                                tarball_files.append( td_tup )
                    # And add output file tuples to the list.
                    for label, filename, _ in test.outputs:
                        output_filepath = os.path.abspath( os.path.join( 'test-data', filename ) )
                        if os.path.exists( output_filepath ):
                            td_tup = ( output_filepath, os.path.join( 'test-data', filename ) )
                            tarball_files.append( td_tup )
            for param in tool.input_params:
                # Check for tool data table definitions.
                if hasattr( param, 'options' ):
                    if hasattr( param.options, 'tool_data_table' ):
                        data_table = param.options.tool_data_table
                        if hasattr( data_table, 'filenames' ):
                            data_table_definitions = []
                            for data_table_filename in data_table.filenames:
                                # FIXME: from_shed_config seems to always be False.
                                if not data_table.filenames[ data_table_filename ][ 'from_shed_config' ]:
                                    tar_file = data_table.filenames[ data_table_filename ][ 'filename' ] + '.sample'
                                    sample_file = os.path.join( data_table.filenames[ data_table_filename ][ 'tool_data_path' ],
                                                                tar_file )
                                    # Use the .sample file, if one exists. If not, skip this data table.
                                    if os.path.exists( sample_file ):
                                        tarfile_path, tarfile_name = os.path.split( tar_file )
                                        tarfile_path = os.path.join( 'tool-data', tarfile_name )
                                        tarball_files.append( ( sample_file, tarfile_path ) )
                                    data_table_definitions.append( data_table.xml_string )
                            if len( data_table_definitions ) > 0:
                                # Put the data table definition XML in a temporary file.
                                table_definition = '<?xml version="1.0" encoding="utf-8"?>\n<tables>\n    %s</tables>'
                                table_definition = table_definition % '\n'.join( data_table_definitions )
                                fd, table_conf = tempfile.mkstemp()
                                os.close( fd )
                                open( table_conf, 'w' ).write( table_definition )
                                tarball_files.append( ( table_conf, os.path.join( 'tool-data', 'tool_data_table_conf.xml.sample' ) ) )
                                temp_files.append( table_conf )
            # Create the tarball.
            fd, tarball_archive = tempfile.mkstemp( suffix='.tgz' )
            os.close( fd )
            tarball = tarfile.open( name=tarball_archive, mode='w:gz' )
            # Add the files from the previously generated list.
            for fspath, tarpath in tarball_files:
                tarball.add( fspath, arcname=tarpath )
            tarball.close()
            # Delete any temporary files that were generated.
            for temp_file in temp_files:
                os.remove( temp_file )
            return tarball_archive, True, None
        return None, False, "An unknown error occurred."

    def reload_tool_by_id( self, tool_id ):
        """
        Attempt to reload the tool identified by 'tool_id', if successful
        replace the old tool.
        """
        if tool_id not in self._tools_by_id:
            message = "No tool with id %s" % escape( tool_id )
            status = 'error'
        else:
            old_tool = self._tools_by_id[ tool_id ]
            new_tool = self.load_tool( old_tool.config_file )
            # The tool may have been installed from a tool shed, so set the tool shed attributes.
            # Since the tool version may have changed, we don't override it here.
            new_tool.id = old_tool.id
            new_tool.guid = old_tool.guid
            new_tool.tool_shed = old_tool.tool_shed
            new_tool.repository_name = old_tool.repository_name
            new_tool.repository_owner = old_tool.repository_owner
            new_tool.installed_changeset_revision = old_tool.installed_changeset_revision
            new_tool.old_id = old_tool.old_id
            # Replace old_tool with new_tool in self._tool_panel
            tool_key = 'tool_' + tool_id
            for key, val in self._tool_panel.items():
                if key == tool_key:
                    self._tool_panel[ key ] = new_tool
                    break
                elif key.startswith( 'section' ):
                    if tool_key in val.elems:
                        self._tool_panel[ key ].elems[ tool_key ] = new_tool
                        break
            # (Re-)Register the reloaded tool, this will handle
            #  _tools_by_id and _tool_versions_by_id
            self.register_tool( new_tool )
            message = "Reloaded the tool:<br/>"
            message += "<b>name:</b> %s<br/>" % escape( old_tool.name )
            message += "<b>id:</b> %s<br/>" % escape( old_tool.id )
            message += "<b>version:</b> %s" % escape( old_tool.version )
            status = 'done'
        return message, status

    def remove_tool_by_id( self, tool_id, remove_from_panel=True ):
        """
        Attempt to remove the tool identified by 'tool_id'. Ignores
        tool lineage - so to remove a tool with potentially multiple
        versions send remove_from_panel=False and handle the logic of
        promoting the next newest version of the tool into the panel
        if needed.
        """
        if tool_id not in self._tools_by_id:
            message = "No tool with id %s" % escape( tool_id )
            status = 'error'
        else:
            tool = self._tools_by_id[ tool_id ]
            del self._tools_by_id[ tool_id ]
            if remove_from_panel:
                tool_key = 'tool_' + tool_id
                for key, val in self._tool_panel.items():
                    if key == tool_key:
                        del self._tool_panel[ key ]
                        break
                    elif key.startswith( 'section' ):
                        if tool_key in val.elems:
                            del self._tool_panel[ key ].elems[ tool_key ]
                            break
                if tool_id in self.data_manager_tools:
                    del self.data_manager_tools[ tool_id ]
            # TODO: do we need to manually remove from the integrated panel here?
            message = "Removed the tool:<br/>"
            message += "<b>name:</b> %s<br/>" % escape( tool.name )
            message += "<b>id:</b> %s<br/>" % escape( tool.id )
            message += "<b>version:</b> %s" % escape( tool.version )
            status = 'done'
        return message, status

    def get_sections( self ):
        for k, v in self._tool_panel.items():
            if isinstance( v, ToolSection ):
                yield (v.id, v.name)

    def find_section_id( self, tool_panel_section_id ):
        """
        Find the section ID referenced by the key or return '' indicating
        no such section id.
        """
        if not tool_panel_section_id:
            tool_panel_section_id = ''
        else:
            if tool_panel_section_id not in self._tool_panel:
                # Hack introduced without comment in a29d54619813d5da992b897557162a360b8d610c-
                # not sure why it is needed.
                fixed_tool_panel_section_id = 'section_%s' % tool_panel_section_id
                if fixed_tool_panel_section_id in self._tool_panel:
                    tool_panel_section_id = fixed_tool_panel_section_id
                else:
                    tool_panel_section_id = ''
        return tool_panel_section_id

    def _load_workflow( self, workflow_id ):
        """
        Return an instance of 'Workflow' identified by `id`,
        which is encoded in the tool panel.
        """
        id = self.app.security.decode_id( workflow_id )
        stored = self.app.model.context.query( self.app.model.StoredWorkflow ).get( id )
        return stored.latest_workflow

    @property
    def sa_session( self ):
        """
        Returns a SQLAlchemy session
        """
        return self.app.model.context

    def tool_panel_contents( self, trans, **kwds ):
        """ Filter tool_panel contents for displaying for user.
        """
        filter_method = self._build_filter_method( trans )
        for _, item_type, elt in self._tool_panel.panel_items_iter():
            elt = filter_method( elt, item_type )
            if elt:
                yield elt

    def to_dict( self, trans, in_panel=True, **kwds ):
        """
        to_dict toolbox.
        """
        if in_panel:
            panel_elts = list( self.tool_panel_contents( trans, **kwds ) )
            # Produce panel.
            rval = []
            kwargs = dict(
                trans=trans,
                link_details=True
            )
            for elt in panel_elts:
                rval.append( elt.to_dict( **kwargs ) )
        else:
            filter_method = self._build_filter_method( trans )
            tools = []
            for id, tool in self._tools_by_id.items():
                tool = filter_method( tool, panel_item_types.TOOL )
                if not tool:
                    continue
                tools.append( tool.to_dict( trans, link_details=True ) )
            rval = tools

        return rval

    def _lineage_in_panel( self, panel_dict, tool=None, tool_lineage=None ):
        """ If tool with same lineage already in panel (or section) - find
        and return it. Otherwise return None.
        """
        if tool_lineage is None:
            assert tool is not None
            if not hasattr( tool, "lineage" ):
                return None
            tool_lineage = tool.lineage
        lineage_tool_versions = tool_lineage.get_versions( reverse=True )
        for lineage_tool_version in lineage_tool_versions:
            lineage_tool = self._tool_from_lineage_version( lineage_tool_version )
            if lineage_tool:
                lineage_id = lineage_tool.id
                if panel_dict.has_tool_with_id( lineage_id ):
                    return panel_dict.get_tool_with_id( lineage_id )
        return None

    def _newer_tool( self, tool1, tool2 ):
        """ Return True if tool1 is considered "newer" given its own lineage
        description.
        """
        if not hasattr( tool1, "lineage" ):
            return True
        lineage_tool_versions = tool1.lineage.get_versions()
        for lineage_tool_version in lineage_tool_versions:
            lineage_tool = self._tool_from_lineage_version( lineage_tool_version )
            if lineage_tool is tool1:
                return False
            if lineage_tool is tool2:
                return True
        return True

    def _tool_from_lineage_version( self, lineage_tool_version ):
        if lineage_tool_version.id_based:
            return self._tools_by_id.get( lineage_tool_version.id, None )
        else:
            return self._tool_versions_by_id.get( lineage_tool_version.id, {} ).get( lineage_tool_version.version, None )

    def _build_filter_method( self, trans ):
        context = Bunch( toolbox=self, trans=trans )
        filters = self._filter_factory.build_filters( trans )
        return lambda element, item_type: _filter_for_panel(element, item_type, filters, context)


def _filter_for_panel( item, item_type, filters, context ):
    """
    Filters tool panel elements so that only those that are compatible
    with provided filters are kept.
    """
    def _apply_filter( filter_item, filter_list ):
        for filter_method in filter_list:
            if not filter_method( context, filter_item ):
                return False
        return True
    if item_type == panel_item_types.TOOL:
        if _apply_filter( item, filters[ 'tool' ] ):
            return item
    elif item_type == panel_item_types.LABEL:
        if _apply_filter( item, filters[ 'label' ] ):
            return item
    elif item_type == panel_item_types.SECTION:
        # Filter section item-by-item. Only show a label if there are
        # non-filtered tools below it.

        if _apply_filter( item, filters[ 'section' ] ):
            cur_label_key = None
            tools_under_label = False
            filtered_elems = item.elems.copy()
            for key, section_item_type, section_item in item.panel_items_iter():
                if section_item_type == panel_item_types.TOOL:
                    # Filter tool.
                    if _apply_filter( section_item, filters[ 'tool' ] ):
                        tools_under_label = True
                    else:
                        del filtered_elems[ key ]
                elif section_item_type == panel_item_types.LABEL:
                    # If there is a label and it does not have tools,
                    # remove it.
                    if ( cur_label_key and not tools_under_label ) or not _apply_filter( section_item, filters[ 'label' ] ):
                        del filtered_elems[ cur_label_key ]

                    # Reset attributes for new label.
                    cur_label_key = key
                    tools_under_label = False

            # Handle last label.
            if cur_label_key and not tools_under_label:
                del filtered_elems[ cur_label_key ]

            # Only return section if there are elements.
            if len( filtered_elems ) != 0:
                copy = item.copy()
                copy.elems = filtered_elems
                return copy

    return None



import pickle


def cmp(x, y):
    # Builtin in Python 2, but not Python 3.
    return (x > y) - (x < y)


class DictionaryObject(object):
    """
    Copyright 2012 "Grim Apps"

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Helper class written by William Grim - grimwm
    Original repo: https://github.com/grimwm/py-dictobj
    Thank you!
    ************************
    A class that has all the functionality of a normal Python dictionary, except
    for the fact it is itself immutable.  It also has the added feature of
    being able to lookup values by using keys as attributes.

    The reason for the class being immutable by default is to help make it a
    little easier to use in multiprocessing situations.  Granted, the underlying
    values themselves are not deeply copied, but the aim is to enforce some
    ensurances of immutability on the container class.

    When using positional arguments, the first argument must always be something
    that would be a valid argument for a dict().  However, a second, optional
    argument may be passed to create a default value when keys are not found.

    Examples:
    >>> d = DictionaryObject({'a':1, 'b':True, 3:'x'})
    >>> print d.a, d.b, d[3]
    1 True x

    >>> d = DictionaryObject((('a',1),('b',2)))
    >>> print d.a, d.b
    1 2

    >>> d = DictionaryObject(a=1, b=True)
    >>> print d
    DictionaryObject({'a': 1, 'b': True})

    >>> d = DictionaryObject({'a':1, 'b':True}, None)
    >>> print d.a, d.b, d.c, d.d
    1 True None None

    >>> d = DictionaryObject({'a':1}, None)
    >>> m = MutableDictionaryObject(d)
    >>> print d == m
    True
    >>> m.a = 0
    >>> print d == m, d < m, d > m, d != m, d <= m, d >= m
    False False True True False True

    >>> import pickle
    >>> m1 = MutableDictionaryObject({'a':1}, None)
    >>> m2 = pickle.loads(pickle.dumps(m1))
    >>> print m1 == m2
    True
    >>> m1.a = 3
    >>> print m1 == m2
    False
    >>> m1.a == 3
    True
    >>> m1['c'] = 5
    >>> m1['c']
    5
    """

    def __init__(self, contents=(), *args, **kwargs):
        """
        Take as input a dictionary-like object and return a DictionaryObject.
        It also makes sure any keys containing dictionaries are also converted
        to DictionaryObjects.
        """
        super(DictionaryObject, self).__init__()
        if isinstance(contents, DictionaryObject):
            self.__dict__.update(pickle.loads(pickle.dumps(contents.__dict__)))
            return

        self.__dict__['_items'] = dict(contents, **kwargs)

        if len(args) > 1:
            raise TypeError("too many arguments")

        # If we have more than one argument passed in, use the second argument
        # as a default value.
        if args:
            try:
                default = type(self)(args[0])
            except:
                default = args[0]
            self.__dict__['_defaultValue'] = default
        else:
            self.__dict__['_defaultValue'] = None
        self.__dict__['_defaultIsSet'] = len(args) > 0

        for k in self._items:
            if isinstance(self._items[k], dict):
                self._items[k] = type(self)(self._items[k])

    def __setstate__(self, dict):
        self.__dict__.update(dict)

    def __getstate__(self):
        return self.__dict__.copy()

    def __getattr__(self, name):
        """
        This is the method that makes all the magic happen.  Search for
        'name' in self._items and return the value if found.  If a default
        value has been set and 'name' is not found in self._items, return it.
        Otherwise raise an AttributeError.

        Example:
          >>> d = DictionaryObject({'keys':[1,2], 'values':3, 'x':1})
          >>> d.keys
          <bound method DictionaryObject.keys of DictionaryObject({'keys': [1, 2], 'x': 1, 'values': 3})>
          >>> d.values
          <bound method DictionaryObject.values of DictionaryObject({'keys': [1, 2], 'x': 1, 'values': 3})>
          >>> d.x
          1
          >>> d['keys']
          [1, 2]
          >>> d['values']
          3
        """
        if name in self._items:
            return self._items[name]
        if self._defaultIsSet:
            return self._defaultValue
        raise AttributeError("'%s' object has no attribute '%s'" % (type(self).__name__, name))

    def __setattr__(self, name, value):
        """
        This class is immutable-by-default.  See MutableDictionaryObject.
        """
        raise AttributeError("'%s' object does not support assignment" % type(self).__name__)

    def __getitem__(self, name):
        return self._items[name]

    def __contains__(self, name):
        return name in self._items

    def __len__(self):
        return len(self._items)

    def __iter__(self):
        return iter(self._items)

    def __repr__(self):
        if self._defaultIsSet:
            params = "%s, %s" % (repr(self._items), self._defaultValue)
        else:
            params = repr(self._items)
        return "%s(%s)" % (type(self).__name__, params)

    def __cmp__(self, rhs):
        if self < rhs:
            return -1
        if self > rhs:
            return 1
        return 0

    def __eq__(self, rhs):
        val = cmp(self._items, rhs._items)
        if 0 == val:
            return 0 == cmp(self._defaultValue, rhs._defaultValue)
        return 0 == val

    def __ne__(self, rhs):
        return not (self == rhs)

    def __lt__(self, rhs):
        val = cmp(self._items, rhs._items)
        if 0 == val:
            return -1 == cmp(self._defaultValue, rhs._defaultValue)
        return -1 == val

    def __le__(self, rhs):
        return self < rhs or self == rhs

    def __gt__(self, rhs):
        return not (self <= rhs)

    def __ge__(self, rhs):
        return self > rhs or self == rhs

    def keys(self):
        return self._items.keys()

    def values(self):
        return self._items.values()

    def asdict(self):
        """
        Copy the data back out of here and into a dict.  Then return it.
        Some libraries may check specifically for dict objects, such
        as the json library; so, this makes it convenient to get the data
        back out.

        >>> import dictobj
        >>> d = {'a':1, 'b':2}
        >>> dictobj.DictionaryObject(d).asdict() == d
        True
        >>> d['c'] = {1:2, 3:4}
        >>> dictobj.DictionaryObject(d).asdict() == d
        True
        """
        items = {}
        for name in self._items:
            value = self._items[name]
        if isinstance(value, DictionaryObject):
            items[name] = value.asdict()
        else:
            items[name] = value
        return items


class MutableDictionaryObject(DictionaryObject):
    """
      Slight enhancement of the DictionaryObject allowing one to add
      attributes easily, in cases where that functionality is wanted.

      Examples:
        >>> d = MutableDictionaryObject({'a':1, 'b':True}, None)
        >>> print d.a, d.b, d.c, d.d
        1 True None None
        >>> d.c = 3
        >>> del d.a
        >>> print d.a, d.b, d.c, d.d
        None True 3 None
    """
    def __setattr__(self, name, value):
        self._items[name] = value

    def __delattr__(self, name):
        del self._items[name]

    __setitem__ = __setattr__
    __delitem__ = __delattr__



import fnmatch
import glob
import os
import re
from ..tools import loader

import sys

import logging
log = logging.getLogger(__name__)

PATH_DOES_NOT_EXIST_ERROR = "Could not load tools from path [%s] - this path does not exist."
PATH_AND_RECURSIVE_ERROR = "Cannot specify a single file and recursive."
LOAD_FAILURE_ERROR = "Failed to load tool with path %s."
TOOL_LOAD_ERROR = object()
TOOL_REGEX = re.compile(r"<tool\s")


def load_exception_handler(path, exc_info):
    log.warn(LOAD_FAILURE_ERROR % path, exc_info=exc_info)


def load_tool_elements_from_path(
    path,
    load_exception_handler=load_exception_handler,
    recursive=False,
    register_load_errors=False,
):
    tool_elements = []
    for file in __find_tool_files(path, recursive=recursive):
        try:
            looks_like_a_tool = __looks_like_a_tool(file)
        except IOError:
            # Some problem reading the tool file, skip.
            continue

        if looks_like_a_tool:
            try:
                tool_elements.append((file, loader.load_tool(file)))
            except Exception:
                exc_info = sys.exc_info()
                load_exception_handler(file, exc_info)
                if register_load_errors:
                    tool_elements.append((file, TOOL_LOAD_ERROR))
    return tool_elements


def is_tool_load_error(obj):
    return obj is TOOL_LOAD_ERROR


def __looks_like_a_tool(path):
    with open(path, "r") as f:
        start_contents = f.read(5 * 1024)
        if TOOL_REGEX.search(start_contents):
            return True
    return False


def __find_tool_files(path, recursive):
    is_file = not os.path.isdir(path)
    if not os.path.exists(path):
        raise Exception(PATH_DOES_NOT_EXIST_ERROR)
    elif is_file and recursive:
        raise Exception(PATH_AND_RECURSIVE_ERROR)
    elif is_file:
        return [os.path.abspath(path)]
    else:
        if not recursive:
            files = glob.glob(path + "/*.xml")
        else:
            files = _find_files(path, "*.xml")
        return map(os.path.abspath, files)


def _find_files(directory, pattern='*'):
    if not os.path.exists(directory):
        raise ValueError("Directory not found {}".format(directory))

    matches = []
    for root, dirnames, filenames in os.walk(directory):
        for filename in filenames:
            full_path = os.path.join(root, filename)
            if fnmatch.filter([full_path], pattern):
                matches.append(os.path.join(root, filename))
    return matches



from abc import (
    ABCMeta,
    abstractmethod
)
import os
import string

from galaxy.util import asbool
from ..deps import docker_util

import logging
log = logging.getLogger(__name__)

DEFAULT_CONTAINER_TYPE = "docker"

LOAD_CACHED_IMAGE_COMMAND_TEMPLATE = '''
python << EOF
import re, tarfile, json, subprocess
t = tarfile.TarFile("${cached_image_file}")
meta_str = t.extractfile('repositories').read()
meta = json.loads(meta_str)
tag, tag_value = meta.items()[0]
rev, rev_value = tag_value.items()[0]
cmd = "${images_cmd}"
proc = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
stdo, stde = proc.communicate()
found = False
for line in stdo.split("\\n"):
    tmp = re.split(r'\s+', line)
    if tmp[0] == tag and tmp[1] == rev and tmp[2] == rev_value:
        found = True
if not found:
    print "Loading image"
    cmd = "cat ${cached_image_file} | ${load_cmd}"
    subprocess.check_call(cmd, shell=True)
EOF
'''


class ContainerFinder(object):

    def __init__(self, app_info):
        self.app_info = app_info
        self.container_registry = ContainerRegistry()

    def find_container(self, tool_info, destination_info, job_info):
        def __destination_container(container_description=None, container_id=None, container_type=None):
            if container_description:
                container_id = container_description.identifier
                container_type = container_description.type
            container = self.__destination_container(
                container_id,
                container_type,
                tool_info,
                destination_info,
                job_info
            )
            return container

        # Is destination forcing Galaxy to use a particular container do it,
        # this is likely kind of a corner case. For instance if deployers
        # do not trust the containers annotated in tools.
        for container_type in CONTAINER_CLASSES.keys():
            container_id = self.__overridden_container_id(container_type, destination_info)
            if container_id:
                container = __destination_container(container_type=container_type, container_id=container_id)
                if container:
                    return container

        # Otherwise lets see if we can find container for the tool.

        # Exact matches first from explicitly listed containers in tools...
        for container_description in tool_info.container_descriptions:
            container = __destination_container(container_description)
            if container:
                return container

        # Implement vague concept of looping through all containers
        # matching requirements. Exact details need to be worked through
        # but hopefully the idea that it sits below find_container somewhere
        # external components to this module don't need to worry about it
        # is good enough.
        container_descriptions = self.container_registry.container_descriptions_for_requirements(tool_info.requirements)
        for container_description in container_descriptions:
            container = __destination_container(container_description)
            if container:
                return container

        # If we still don't have a container, check to see if any container
        # types define a default container id and use that.
        for container_type in CONTAINER_CLASSES.keys():
            container_id = self.__default_container_id(container_type, destination_info)
            if container_id:
                container = __destination_container(container_type=container_type, container_id=container_id)
                if container:
                    return container

        return NULL_CONTAINER

    def __overridden_container_id(self, container_type, destination_info):
        if not self.__container_type_enabled(container_type, destination_info):
            return None
        return destination_info.get("%s_container_id_override" % container_type)

    def __default_container_id(self, container_type, destination_info):
        if not self.__container_type_enabled(container_type, destination_info):
            return None
        return destination_info.get("%s_default_container_id" % container_type)

    def __destination_container(self, container_id, container_type, tool_info, destination_info, job_info):
        # TODO: ensure destination_info is dict-like
        if not self.__container_type_enabled(container_type, destination_info):
            return NULL_CONTAINER

        # TODO: Right now this assumes all containers available when a
        # container type is - there should be more thought put into this.
        # Checking which are availalbe - settings policies for what can be
        # auto-fetched, etc....
        return CONTAINER_CLASSES[container_type](container_id, self.app_info, tool_info, destination_info, job_info)

    def __container_type_enabled(self, container_type, destination_info):
        return asbool(destination_info.get("%s_enabled" % container_type, False))


class NullContainerFinder(object):

    def find_container(self, tool_info, destination_info, job_info):
        return []


class ContainerRegistry():

    def __init__(self):
        pass

    def container_descriptions_for_requirements(self, requirements):
        # Return lists of containers that would match requirements...
        return []


class AppInfo(object):

    def __init__(
        self,
        galaxy_root_dir=None,
        default_file_path=None,
        outputs_to_working_directory=False,
        container_image_cache_path=None,
    ):
        self.galaxy_root_dir = galaxy_root_dir
        self.default_file_path = default_file_path
        # TODO: Vary default value for docker_volumes based on this...
        self.outputs_to_working_directory = outputs_to_working_directory
        self.container_image_cache_path = container_image_cache_path


class ToolInfo(object):
    # TODO: Introduce tool XML syntax to annotate the optional environment
    # variables they can consume (e.g. JVM options, license keys, etc..)
    # and add these to env_path_through

    def __init__(self, container_descriptions=[], requirements=[]):
        self.container_descriptions = container_descriptions
        self.requirements = requirements
        self.env_pass_through = ["GALAXY_SLOTS"]


class JobInfo(object):

    def __init__(self, working_directory, tool_directory, job_directory):
        self.working_directory = working_directory
        self.job_directory = job_directory
        # Tool files may be remote staged - so this is unintuitively a property
        # of the job not of the tool.
        self.tool_directory = tool_directory


class Container( object ):
    __metaclass__ = ABCMeta

    def __init__(self, container_id, app_info, tool_info, destination_info, job_info):
        self.container_id = container_id
        self.app_info = app_info
        self.tool_info = tool_info
        self.destination_info = destination_info
        self.job_info = job_info

    @abstractmethod
    def containerize_command(self, command):
        """
        Use destination supplied container configuration parameters,
        container_id, and command to build a new command that runs
        input command in container.
        """


class DockerContainer(Container):

    def containerize_command(self, command):
        def prop(name, default):
            destination_name = "docker_%s" % name
            return self.destination_info.get(destination_name, default)

        env_directives = []
        for pass_through_var in self.tool_info.env_pass_through:
            env_directives.append('"%s=$%s"' % (pass_through_var, pass_through_var))

        # Allow destinations to explicitly set environment variables just for
        # docker container. Better approach is to set for destination and then
        # pass through only what tool needs however. (See todo in ToolInfo.)
        for key, value in self.destination_info.iteritems():
            if key.startswith("docker_env_"):
                env = key[len("docker_env_"):]
                env_directives.append('"%s=%s"' % (env, value))

        working_directory = self.job_info.working_directory
        if not working_directory:
            raise Exception("Cannot containerize command [%s] without defined working directory." % working_directory)

        volumes_raw = self.__expand_str(self.destination_info.get("docker_volumes", "$defaults"))
        # TODO: Remove redundant volumes...
        volumes = docker_util.DockerVolume.volumes_from_str(volumes_raw)
        volumes_from = self.destination_info.get("docker_volumes_from", docker_util.DEFAULT_VOLUMES_FROM)

        docker_host_props = dict(
            docker_cmd=prop("cmd", docker_util.DEFAULT_DOCKER_COMMAND),
            sudo=asbool(prop("sudo", docker_util.DEFAULT_SUDO)),
            sudo_cmd=prop("sudo_cmd", docker_util.DEFAULT_SUDO_COMMAND),
            host=prop("host", docker_util.DEFAULT_HOST),
        )

        cached_image_file = self.__get_cached_image_file()
        if not cached_image_file:
            # TODO: Add option to cache it once here and create cached_image_file.
            cache_command = docker_util.build_docker_cache_command(self.container_id, **docker_host_props)
        else:
            cache_command = self.__cache_from_file_command(cached_image_file, docker_host_props)
        run_command = docker_util.build_docker_run_command(
            command,
            self.container_id,
            volumes=volumes,
            volumes_from=volumes_from,
            env_directives=env_directives,
            working_directory=working_directory,
            net=prop("net", "none"),  # By default, docker instance has networking disabled
            auto_rm=asbool(prop("auto_rm", docker_util.DEFAULT_AUTO_REMOVE)),
            set_user=prop("set_user", docker_util.DEFAULT_SET_USER),
            run_extra_arguments=prop("run_extra_arguments", docker_util.DEFAULT_RUN_EXTRA_ARGUMENTS),
            **docker_host_props
        )
        return "%s\n%s" % (cache_command, run_command)

    def __cache_from_file_command(self, cached_image_file, docker_host_props):
        images_cmd = docker_util.build_docker_images_command(truncate=False, **docker_host_props)
        load_cmd = docker_util.build_docker_load_command(**docker_host_props)

        return string.Template(LOAD_CACHED_IMAGE_COMMAND_TEMPLATE).safe_substitute(
            cached_image_file=cached_image_file,
            images_cmd=images_cmd,
            load_cmd=load_cmd
        )

    def __get_cached_image_file(self):
        container_id = self.container_id
        cache_directory = os.path.abspath(self.__get_destination_overridable_property("container_image_cache_path"))
        cache_path = docker_cache_path(cache_directory, container_id)
        return cache_path if os.path.exists(cache_path) else None

    def __get_destination_overridable_property(self, name):
        prop_name = "docker_%s" % name
        if prop_name in self.destination_info:
            return self.destination_info[prop_name]
        else:
            return getattr(self.app_info, name)

    def __expand_str(self, value):
        if not value:
            return value

        template = string.Template(value)
        variables = dict()

        def add_var(name, value):
            if value:
                variables[name] = os.path.abspath(value)

        add_var("working_directory", self.job_info.working_directory)
        add_var("job_directory", self.job_info.job_directory)
        add_var("tool_directory", self.job_info.tool_directory)
        add_var("galaxy_root", self.app_info.galaxy_root_dir)
        add_var("default_file_path", self.app_info.default_file_path)

        if self.job_info.job_directory:
            # We have a job directory, so everything needed (excluding index
            # files) should be available in job_directory...
            defaults = "$job_directory:ro,$tool_directory:ro,$job_directory/outputs:rw,$working_directory:rw"
        elif self.app_info.outputs_to_working_directory:
            # Should need default_file_path (which is a course estimate given
            # object stores anyway.
            defaults = "$galaxy_root:ro,$tool_directory:ro,$working_directory:rw,$default_file_path:ro"
        else:
            defaults = "$galaxy_root:ro,$tool_directory:ro,$working_directory:rw,$default_file_path:rw"

        # Define $defaults that can easily be extended with external library and
        # index data without deployer worrying about above details.
        variables["defaults"] = string.Template(defaults).safe_substitute(variables)

        return template.safe_substitute(variables)


def docker_cache_path(cache_directory, container_id):
    file_container_id = container_id.replace("/", "_slash_")
    cache_file_name = "docker_%s.tar" % file_container_id
    return os.path.join(cache_directory, cache_file_name)


CONTAINER_CLASSES = dict(
    docker=DockerContainer,
)


class NullContainer(object):

    def __init__(self):
        pass

    def __nonzero__(self):
        return False


NULL_CONTAINER = NullContainer()



from os.path import join, islink, realpath, basename, exists

from ..resolvers import DependencyResolver, INDETERMINATE_DEPENDENCY, Dependency
from .resolver_mixins import UsesToolDependencyDirMixin

import logging
log = logging.getLogger( __name__ )


class GalaxyPackageDependencyResolver(DependencyResolver, UsesToolDependencyDirMixin):
    resolver_type = "galaxy_packages"

    def __init__(self, dependency_manager, **kwds):
        # Galaxy tool shed requires explicit versions on XML elements,
        # this in inconvient for testing or Galaxy instances not utilizing
        # the tool shed so allow a fallback version of the Galaxy package
        # resolver that will just grab 'default' version of exact version
        # unavailable.
        self.versionless = str(kwds.get('versionless', "false")).lower() == "true"
        self._init_base_path( dependency_manager, **kwds )

    def resolve( self, name, version, type, **kwds ):
        """
        Attempt to find a dependency named `name` at version `version`. If version is None, return the "default" version as determined using a
        symbolic link (if found). Returns a triple of: env_script, base_path, real_version
        """
        if version is None or self.versionless:
            return self._find_dep_default( name, type=type, **kwds )
        else:
            return self._find_dep_versioned( name, version, type=type, **kwds )

    def _find_dep_versioned( self, name, version, type='package', **kwds ):
        base_path = self.base_path
        path = join( base_path, name, version )
        return self._galaxy_package_dep(path, version)

    def _find_dep_default( self, name, type='package', **kwds ):
        base_path = self.base_path
        path = join( base_path, name, 'default' )
        if islink( path ):
            real_path = realpath( path )
            real_version = basename( real_path )
            return self._galaxy_package_dep(real_path, real_version)
        else:
            return INDETERMINATE_DEPENDENCY

    def _galaxy_package_dep( self, path, version ):
        script = join( path, 'env.sh' )
        if exists( script ):
            return GalaxyPackageDependency(script, path, version)
        elif exists( join( path, 'bin' ) ):
            return GalaxyPackageDependency(None, path, version)
        return INDETERMINATE_DEPENDENCY


class GalaxyPackageDependency(Dependency):

    def __init__( self, script, path, version ):
        self.script = script
        self.path = path
        self.version = version

    def shell_commands( self, requirement ):
        base_path = self.path
        if self.script is None and base_path is None:
            log.warn( "Failed to resolve dependency on '%s', ignoring", requirement.name )
            commands = None
        elif requirement.type == 'package' and self.script is None:
            commands = 'PACKAGE_BASE=%s; export PACKAGE_BASE; PATH="%s/bin:$PATH"; export PATH' % ( base_path, base_path )
        else:
            commands = 'PACKAGE_BASE=%s; export PACKAGE_BASE; . %s' % ( base_path, self.script )
        return commands

__all__ = ['GalaxyPackageDependencyResolver', 'GalaxyPackageDependency']





class JobMetricFormatter( object ):
    """ Format job metric key-value pairs for human consumption in Web UI. """

    def format( self, key, value ):
        return ( str( key ), str( value ) )


# Formatting utilities

def seconds_to_str( value ):
    if value < 60:
        return "%s seconds" % value
    elif value < 3600:
        return "%s minutes" % ( value / 60 )
    else:
        return "%s hours and %s minutes" % ( value / 3600, ( value % 3600 ) / 60 )



"""
Here for compat. with objectstore.
"""


class ObjectNotFound(Exception):
    """ Accessed object was not found """
    pass


class ObjectInvalid(Exception):
    """ Accessed object store ID is invalid """
    pass



from ..lint_util import is_datasource


# Misspelled so as not be picked up by nosetests.
def lint_tsts(tool_xml, lint_ctx):
    tests = tool_xml.findall("./tests/test")
    datasource = is_datasource(tool_xml)

    if not tests and not datasource:
        lint_ctx.warn("No tests found, most tools should define test cases.")
    elif datasource:
        lint_ctx.info("No tests found, that should be OK for data_sources.")

    num_valid_tests = 0
    for test in tests:
        has_test = False
        if "expect_failure" in test.attrib or "expect_exit_code" in test.attrib:
            has_test = True
        if len(test.findall("assert_stdout")) > 0:
            has_test = True
        if len(test.findall("assert_stdout")) > 0:
            has_test = True

        outputs = test.findall("output") + test.findall("output_collection")
        if len(outputs) > 0:
            has_test = True
        if not has_test:
            lint_ctx.warn("No outputs or expectations defined for tests, this test is likely invalid.")
        else:
            num_valid_tests += 1

    if num_valid_tests or datasource:
        lint_ctx.valid("%d test(s) found.", num_valid_tests)
    else:
        lint_ctx.warn("No valid test(s) found.")



import re


def lint_top_level(tree, lint_ctx):
    root = tree.getroot()
    if "version" not in root.attrib:
        lint_ctx.error("Tool does not define a version attribute.")
    else:
        lint_ctx.valid("Tool defines a version.")

    if "name" not in root.attrib:
        lint_ctx.error("Tool does not define a name attribute.")
    else:
        lint_ctx.valid("Tool defines a name.")

    if "id" not in root.attrib:
        lint_ctx.error("Tool does not define an id attribute.")
    else:
        lint_ctx.valid("Tool defines an id name.")

    id = root.attrib["id"]
    if re.search(r"\s", id):
        lint_ctx.warn("Tool id contains a space - this is discouraged.")
